package org.droid3d.io;

import java.io.BufferedReader;
import java.io.IOException;
import java.io.InputStreamReader;
import java.net.URL;
import java.util.HashMap;
import java.util.Map;
import java.util.StringTokenizer;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import org.apache.log4j.Logger;
import org.droid3d.math.Float2;
import org.droid3d.renderer.Bitmap;
import org.droid3d.renderer.Font;
import org.droid3d.renderer.Font.Glyph;
import org.droid3d.renderer.Texture;
import org.droid3d.utils.FileNameUtils;

public class BMFontLoader implements FontLoader {

	private static final Logger	LOG	= Logger.getLogger(BMFontLoader.class);

	private static Map<String, Integer> readLine(String line) {
		Map<String, Integer> properties = new HashMap<String, Integer>();
		StringTokenizer tokenizer = new StringTokenizer(line);
		while (tokenizer.hasMoreTokens()) {
			String token = tokenizer.nextToken();
			String[] property = token.split("=");
			properties.put(property[0], Integer.valueOf(property[1]));
		}
		return properties;
	}

	@Override
	public Font load(URL url) throws IOException {
		int lineHeight = 0;
		int base = 0;
		Texture texture = null;
		Map<Character, Glyph> glyphs = new HashMap<Character, Glyph>();

		BufferedReader reader = new BufferedReader(new InputStreamReader(url.openStream()));
		String line = null;
		while ((line = reader.readLine()) != null) {
			if (line.startsWith("common")) {
				Map<String, Integer> commonProperties = readLine(line.substring(7));
				Integer value = commonProperties.get("lineHeight");
				if (value == null) {
					LOG.error("Error loading font - Could not read line height");
					return null;
				}
				lineHeight = value.intValue();

				value = commonProperties.get("base");
				if (value == null) {
					LOG.error("Error loading font - Could not read base line");
					return null;
				}
				base = value.intValue();
			} else if (line.startsWith("page")) {
				Pattern p = Pattern.compile(".*?file=\\\"(.*?)\\\".*?");
				Matcher m = p.matcher(line);
				if (!m.matches()) {
					LOG.error("Error loading font - Could not read texture filename");
					return null;
				}
				URL pageFile = new URL(FileNameUtils.getDirectory(url.toString()) + m.group(1));
				Bitmap bitmap = BitmapFactory.getInstance().load(pageFile);
				if (bitmap == null) {
					return null;
				}
				texture = new Texture(bitmap);
			} else if (line.startsWith("char ")) {
				Map<String, Integer> glyphProperties = readLine(line.substring(5));
				Character c = Character.valueOf((char) glyphProperties.get("id").intValue());

				Font.Glyph glyph = new Font.Glyph();
				glyph.x = glyphProperties.get("x");
				glyph.y = glyphProperties.get("y");
				glyph.width = glyphProperties.get("width");
				glyph.height = glyphProperties.get("height");
				glyph.offset[0] = glyphProperties.get("xoffset");
				glyph.offset[1] = glyphProperties.get("yoffset");
				glyph.xadvance = glyphProperties.get("xadvance");

				glyphs.put(c, glyph);
			}

		}

		float width = texture.getWidth();
		float height = texture.getHeight();

		for (Font.Glyph glyph : glyphs.values()) {
			// @formatter:off
				glyph.uvs[0] = new Float2((glyph.x + glyph.width) / width,  glyph.y / height);
				glyph.uvs[1] = new Float2(glyph.x / width, glyph.y / height);
				glyph.uvs[2] = new Float2(glyph.x / width, (glyph.y + glyph.height) / height);
				glyph.uvs[3] = new Float2((glyph.x + glyph.width) / width, (glyph.y + glyph.height) / height);
				// @formatter:on
		}

		return new Font(lineHeight, base, texture, glyphs);
	}

}
