package org.droid3d.io;

import java.io.ByteArrayOutputStream;
import java.io.DataInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.nio.ByteBuffer;
import java.util.zip.InflaterOutputStream;

import org.apache.log4j.Logger;
import org.droid3d.renderer.Bitmap;
import org.droid3d.utils.Type;

import com.google.common.base.Charsets;

public class PNGDecoder implements BitmapDecoder {
	private static class PNGDataStream extends OutputStream {
		/**
		 * size of a row in bytes
		 */
		private int						mRowSize;

		/**
		 * size of a scan line in bytes
		 */
		private int						mScanLineSize;

		/**
		 * size of a pixel in bytes
		 */
		private int						mPixelSize;
		private int						mFilter		= 0;
		/**
		 * current scan line
		 */
		private byte[]					mCur;

		/**
		 * previous scan line
		 */
		private byte[]					mPrv;

		private ByteArrayOutputStream	mData		= new ByteArrayOutputStream();

		/**
		 * number of bytes consumed
		 */
		private int						mConsumed	= 0;
		private int						mCol		= 0;

		public PNGDataStream(int pixelSize, int imgWidth) {
			mPixelSize = pixelSize;
			mRowSize = mPixelSize * imgWidth;
			mScanLineSize = mRowSize + 1;

			mCur = new byte[mRowSize];
			mPrv = new byte[mRowSize];
		}

		private int defilter(byte v) {
			int x = 0xff & v;
			int a = mCol < mPixelSize ? 0 : 0xff & mCur[mCol - mPixelSize];
			int b = 0xff & mPrv[mCol];
			int c = mCol < mPixelSize ? 0 : 0xff & mPrv[mCol - mPixelSize];

			switch (mFilter) {
			case 0: // None
				return x;
			case 1: // Sub
				return x + a;
			case 2: // Up
				return x + b;
			case 3: // Average
				return x + ((a + b) / 2);
			case 4: // Paeth
				int p = a + b - c;
				int pa = p - a;
				if (pa < 0)
					pa = -pa;

				int pb = p - b;
				if (pb < 0)
					pb = -pb;

				int pc = p - c;
				if (pc < 0)
					pc = -pc;

				if (pa <= pb && pa <= pc) {
					return x + a;
				} else if (pb <= pc) {
					return x + b;
				} else {
					return x + c;
				}
			default:
				throw new UnsupportedOperationException("Unsupported filter type " + mFilter);
			}
		}

		public byte[] getData() {
			return mData.toByteArray();
		}

		@Override
		public void write(int b) throws IOException {
			if (mConsumed % mScanLineSize == 0) {
				// first byte of a scan line is a filter byte
				mFilter = 0xff & b;
			} else {
				mCur[mCol] = (byte) defilter((byte) b);
				mCol++;
				if (mCol == mRowSize) {
					mData.write(mCur);
					System.arraycopy(mCur, 0, mPrv, 0, mRowSize);
					mCol = 0;
				}

			}
			mConsumed++;
		}

	}

	private static class PNGInfo {
		int	width;
		int	height;
		int	bitsPerChannel;
		int	numChannels;
		int	colorType;
		int	compression;
		int	filter;
		int	interlace;
	}

	private static final String	TYPE_IHDR		= "IHDR";
	private static final String	TYPE_IDAT		= "IDAT";
	private static final String	TYPE_IEND		= "IEND";

	private static final int	COLOR_TYPE_RGB	= 2;
	private static final int	COLOR_TYPE_RGBA	= 6;

	private static final Logger	LOG				= Logger.getLogger(PNGDecoder.class);

	@Override
	public Bitmap decode(InputStream in) throws IOException {
		Bitmap bitmap = null;
		DataInputStream dataIn = new DataInputStream(in);

		long signature = dataIn.readLong();
		if (signature != 0x89504e470d0a1a0aL)
			throw new IOException("Signature mismatch");

		PNGInfo info = getInfo(dataIn);
		switch (info.colorType) {
		case COLOR_TYPE_RGB:
			info.numChannels = 3;
			break;
		case COLOR_TYPE_RGBA:
			info.numChannels = 4;
			break;
		default:
			throw new UnsupportedOperationException("Unsupported color type " + info.colorType);
		}

		if (info.bitsPerChannel != 8) {
			throw new UnsupportedOperationException("Unsupported bpp " + info.bitsPerChannel);
		}

		if (info.interlace != 0) {
			throw new UnsupportedOperationException("Unsupported interlace mode " + info.interlace);
		}

		if (info.filter != 0) {
			throw new UnsupportedOperationException("Unsupported filter mode " + info.filter);
		}

		if (info.compression != 0) {
			throw new UnsupportedOperationException("Unsupported compression mode " + info.compression);
		}

		int pixelSize = info.numChannels * info.bitsPerChannel / 8;
		PNGDataStream dataStream = new PNGDataStream(pixelSize, info.width);
		InflaterOutputStream dataOut = new InflaterOutputStream(dataStream);
		String type = TYPE_IEND;
		do {
			int length = dataIn.readInt();

			byte[] typeBytes = new byte[4];
			dataIn.readFully(typeBytes);
			type = new String(typeBytes, Charsets.UTF_8);

			byte[] data = new byte[length];
			dataIn.readFully(data);

			long crc = dataIn.readInt() & 0x00000000ffffffffl;

			if (TYPE_IDAT.equals(type)) {
				dataOut.write(data);
			}
		} while (!TYPE_IEND.equals(type));
		dataOut.close();

		byte[] data = dataStream.getData();
		Bitmap.Builder builder = new Bitmap.Builder();
		builder.setData(data);
		builder.setWidth(info.width);
		builder.setHeight(info.height);
		builder.setFormat(info.colorType == COLOR_TYPE_RGB ? Bitmap.Format.RGB : Bitmap.Format.RGBA);
		builder.setDataType(Type.UNSIGNED_BYTE);
		bitmap = builder.build();

		return bitmap;
	}

	private PNGInfo getInfo(DataInputStream in) throws IOException {
		PNGInfo info = new PNGInfo();

		int length = in.readInt();

		byte[] type = new byte[4];
		in.readFully(type);

		byte[] data = new byte[length];
		in.readFully(data);
		ByteBuffer bb = ByteBuffer.wrap(data);

		info.width = bb.getInt();
		info.height = bb.getInt();
		info.bitsPerChannel = 0xff & bb.get();
		info.colorType = 0xff & bb.get();
		info.compression = 0xff & bb.get();
		info.filter = 0xff & bb.get();
		info.interlace = 0xff & bb.get();

		int crc = in.readInt();
		return info;
	}

}
