package org.droid3d.io;

import static com.google.common.base.Preconditions.checkNotNull;

import java.io.BufferedReader;
import java.io.IOException;
import java.io.InputStreamReader;
import java.net.URL;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;

import org.apache.log4j.Logger;
import org.droid3d.io.MaterialFactory.MaterialLoader;
import org.droid3d.renderer.Color;
import org.droid3d.renderer.Material;

public class WavefrontMaterialLoader implements MaterialLoader {
	private static final Logger LOG = Logger
			.getLogger(WavefrontMaterialLoader.class);

	private interface PropertySetter {
		public enum ColorProperty {
			AMBIENT, DIFFUSE, SPECULAR;
		}

		public enum FloatProperty {
			SHININESS, TRANSPARENCY;
		}

		public void set(Material material, String string);
	}

	private HashMap<String, PropertySetter> mSetter = new HashMap<String, PropertySetter>();

	private static class ColorPropertySetter implements PropertySetter {
		private final ColorProperty mProperty;

		public ColorPropertySetter(ColorProperty property) {
			mProperty = property;
		}

		@Override
		public void set(Material material, String string) {
			Color color = colorFromString(string);
			switch (mProperty) {
			case AMBIENT:
				material.setAmbientColor(color);
				break;
			case DIFFUSE:
				material.setDiffuseColor(color);
				break;
			case SPECULAR:
				material.setSpecularColor(color);
				break;
			}
		}
	}

	private static class FloatPropertySetter implements PropertySetter {
		private FloatProperty mProperty;

		public FloatPropertySetter(FloatProperty property) {
			mProperty = property;
		}

		@Override
		public void set(Material material, String string) {
			float value = Float.parseFloat(string);
			switch (mProperty) {
			case SHININESS:
				material.setShininess(value);
				break;
			case TRANSPARENCY:
				material.setTranspanrency(value);
				break;
			}
		}

	}

	public WavefrontMaterialLoader() {
		mSetter.put("Ka", new ColorPropertySetter(
				PropertySetter.ColorProperty.AMBIENT));
		mSetter.put("Kd", new ColorPropertySetter(
				PropertySetter.ColorProperty.DIFFUSE));
		mSetter.put("Ks", new ColorPropertySetter(
				PropertySetter.ColorProperty.SPECULAR));
		mSetter.put("Ns", new FloatPropertySetter(
				PropertySetter.FloatProperty.SHININESS));
		FloatPropertySetter transparencySetter = new FloatPropertySetter(
				PropertySetter.FloatProperty.TRANSPARENCY);
		mSetter.put("Tr", transparencySetter);
		mSetter.put("d", transparencySetter);
	}

	@Override
	public List<Material> load(URL url) throws IOException {
		checkNotNull(url);

		LOG.debug("Loading material library from " + url.toString());

		BufferedReader reader = null;
		List<Material> materials = new ArrayList<Material>();
		try {
			reader = new BufferedReader(new InputStreamReader(url.openStream()));
			String line = null;
			Material currentMaterial = null;
			int lineNumber = 0;
			while ((line = reader.readLine()) != null) {
				lineNumber++;
				line = line.trim();

				if (line.isEmpty() || line.startsWith("#")) {
					continue;
				}

				int idx = line.indexOf(' ');
				if (idx == -1) {
					throw new RuntimeException(
							"Error loading material: syntax error ('" + line
									+ "', line " + lineNumber + ")");
				}

				String command = line.substring(0, idx);
				String arguments = line.substring(idx + 1);

				if (command.equals("newmtl")) {
					if (currentMaterial != null) {
						materials.add(currentMaterial);
					}

					currentMaterial = new Material();
					currentMaterial.setName(arguments.trim());
				} else if (currentMaterial == null) {
					throw new RuntimeException(
							"Error loading material: missing newmtl statement");
				} else {
					PropertySetter setter = mSetter.get(command);
					if (setter == null) {
						LOG.warn("Unsupported command " + command);
						continue;
					} else {
						setter.set(currentMaterial, arguments);
					}
				}

			}

			if (currentMaterial != null) {
				materials.add(currentMaterial);
			}

		} finally {
			if (reader != null)
				reader.close();
		}

		LOG.debug("Loaded " + materials.size() + " material(s)");
		return materials;
	}

	private static Color colorFromString(String string) {
		String[] values = string.split(" ");
		if (values.length != 3) {
			throw new RuntimeException("Cannot create color from string '"
					+ string + "'");
		}

		float red = Float.parseFloat(values[0].trim());
		float green = Float.parseFloat(values[1].trim());
		float blue = Float.parseFloat(values[2].trim());

		return new Color(red, green, blue);
	}
}
