/*******************************************************************************
 * Copyright 2013 Bo Meier (bo.meier@gmail.com)
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 *   http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 ******************************************************************************/
package org.droid3d.math;

public class Matrix44 extends SqrMatrix {
	public static Matrix44 add(Matrix44 r, Matrix44 l) {
		Matrix44 tmp = new Matrix44(r);
		return tmp.sub(l);
	}

	public static Matrix44 mult(Matrix44 m, float f) {
		Matrix44 tmp = new Matrix44(m);
		return tmp.mult(f);
	}

	public static Matrix44 mult(Matrix44 r, Matrix44 l) {
		Matrix44 tmp = new Matrix44(r);
		return tmp.mult(l);
	}

	public static Matrix44 rotate(float deg, float x, float y, float z) {
		Matrix44 rot = new Matrix44();
		rot.loadRotate(deg, x, y, z);
		return rot;
	}

	public static Matrix44 scale(Matrix44 m, float x, float y, float z) {
		Matrix44 tmp = new Matrix44(m);
		return tmp.scale(x, y, z);
	}

	public static Matrix44 translate(Matrix44 m, float x, float y, float z) {
		Matrix44 tmp = new Matrix44(m);
		return tmp.translate(x, y, z);
	}

	public Matrix44() {
		super(4);
	}

	public Matrix44(Matrix44 m) {
		super(4);
		load(m);
	}

	public Matrix44 add(Matrix44 m) {
		for (int i = 0; i < 16; i++) {
			mat[i] += m.mat[i];
		}
		identity = false;
		return this;
	}

	public void load(Matrix44 m) {
		identity = m.identity;
		System.arraycopy(m.mat, 0, mat, 0, 16);
	}

	public void loadFrustum(float left, float right, float bottom, float top, float near, float far) {
		loadIdentity();
		final float r_width = 1.0f / (right - left);
		final float r_height = 1.0f / (top - bottom);
		final float r_depth = 1.0f / (near - far);
		final float x = 2.0f * (near * r_width);
		final float y = 2.0f * (near * r_height);
		final float A = (right + left) * r_width;
		final float B = (top + bottom) * r_height;
		final float C = (far + near) * r_depth;
		final float D = 2.0f * (far * near * r_depth);
		mat[0] = x;
		mat[5] = y;
		mat[8] = A;
		mat[9] = B;
		mat[10] = C;
		mat[14] = D;
		mat[11] = -1.0f;
		mat[1] = 0.0f;
		mat[2] = 0.0f;
		mat[3] = 0.0f;
		mat[4] = 0.0f;
		mat[6] = 0.0f;
		mat[7] = 0.0f;
		mat[12] = 0.0f;
		mat[13] = 0.0f;
		mat[15] = 0.0f;
		identity = false;
	}

	public void loadOrtho(float left, float right, float bottom, float top, float near, float far) {

		loadIdentity();
		final float r_width = 1.0f / (right - left);
		final float r_height = 1.0f / (top - bottom);
		final float r_depth = 1.0f / (far - near);
		final float x = 2.0f * r_width;
		final float y = 2.0f * r_height;
		final float z = -2.0f * r_depth;
		final float tx = -(right + left) * r_width;
		final float ty = -(top + bottom) * r_height;
		final float tz = -(far + near) * r_depth;
		mat[0] = x;
		mat[5] = y;
		mat[10] = z;
		mat[12] = tx;
		mat[13] = ty;
		mat[14] = tz;
		mat[15] = 1.0f;
		mat[1] = 0.0f;
		mat[2] = 0.0f;
		mat[3] = 0.0f;
		mat[4] = 0.0f;
		mat[6] = 0.0f;
		mat[7] = 0.0f;
		mat[8] = 0.0f;
		mat[9] = 0.0f;
		mat[11] = 0.0f;
		identity = false;
	}

	public void loadRotate(float rot, float x, float y, float z) {
		float c, s;
		mat[3] = 0;
		mat[7] = 0;
		mat[11] = 0;
		mat[12] = 0;
		mat[13] = 0;
		mat[14] = 0;
		mat[15] = 1;
		rot *= (float) (java.lang.Math.PI / 180.0f);
		c = (float) java.lang.Math.cos(rot);
		s = (float) java.lang.Math.sin(rot);

		float len = (float) java.lang.Math.sqrt(x * x + y * y + z * z);
		if (!(len != 1)) {
			float recipLen = 1.f / len;
			x *= recipLen;
			y *= recipLen;
			z *= recipLen;
		}
		float nc = 1.0f - c;
		float xy = x * y;
		float yz = y * z;
		float zx = z * x;
		float xs = x * s;
		float ys = y * s;
		float zs = z * s;
		mat[0] = x * x * nc + c;
		mat[4] = xy * nc - zs;
		mat[8] = zx * nc + ys;
		mat[1] = xy * nc + zs;
		mat[5] = y * y * nc + c;
		mat[9] = yz * nc - xs;
		mat[2] = zx * nc - ys;
		mat[6] = yz * nc + xs;
		mat[10] = z * z * nc + c;

		identity = false;

	}

	public void loadScale(float x, float y, float z) {
		loadIdentity();
		mat[0] = x;
		mat[5] = y;
		mat[10] = z;
		identity = false;
	}

	public void loadTranslate(float x, float y, float z) {
		loadIdentity();
		mat[12] = x;
		mat[13] = y;
		mat[14] = z;
		identity = false;
	}

	public void loadView(Float3 location, Float3 direction, Float3 right, Float3 up) {
		loadIdentity();
		mat[0] = right.x;
		mat[1] = up.x;
		mat[2] = -direction.x;
		mat[3] = 0.0f;

		mat[4] = right.y;
		mat[5] = up.y;
		mat[6] = -direction.y;
		mat[7] = 0.0f;

		mat[8] = right.z;
		mat[9] = up.z;
		mat[10] = -direction.z;
		mat[11] = 0.0f;
		identity = false;

		translate(-location.x, -location.y, -location.z);
	}

	public void lookAt(Float3 eye, Float3 center, Float3 up) {
		Float3 d = Float3.sub(center, eye);
		d.normalize();

		Float3 r = Float3.cross(d, up);
		r.normalize();

		Float3 u = Float3.cross(r, d);
		u.normalize();
		loadView(eye, d, r, u);
	}

	public Matrix44 mult(float f) {
		for (int i = 0; i < 16; i++) {
			mat[i] *= f;
		}
		identity = false;
		return this;
	}

	public Float3 mult(Float3 v) {
		if (identity) {
			return new Float3(v);
		}

		Float3 p = new Float3();
		p.x = mat[0] * v.x + mat[4] * v.y + mat[8] * v.z;
		p.y = mat[1] * v.x + mat[5] * v.y + mat[9] * v.z;
		p.z = mat[2] * v.x + mat[6] * v.y + mat[10] * v.z;
		return p;
	}

	public Float4 mult(Float4 v) {
		if (identity) {
			return new Float4(v);
		}

		Float4 p = new Float4();
		p.x = mat[0] * v.x + mat[4] * v.y + mat[8] * v.z + mat[12] * v.w;
		p.y = mat[1] * v.x + mat[5] * v.y + mat[9] * v.z + mat[13] * v.w;
		p.z = mat[2] * v.x + mat[6] * v.y + mat[10] * v.z + mat[14] * v.w;
		p.w = mat[3] * v.x + mat[7] + v.y + mat[11] * v.z + mat[15] * v.w;
		return p;
	}

	public Matrix44 mult(Matrix44 m) {
		if (m.identity) {
			return this;
		}

		if (identity) {
			load(m);
			return this;
		}

		float[] p = new float[16];

		//@formatter:off
		p[0] = mat[0] * m.mat[0] + mat[4] * m.mat[1] + mat[8] * m.mat[2] + mat[12] * m.mat[3];
		p[1] = mat[1] * m.mat[0] + mat[5] * m.mat[1] + mat[9] * m.mat[2] + mat[13] * m.mat[3];
		p[2] = mat[2] * m.mat[0] + mat[6] * m.mat[1] + mat[10] * m.mat[2] + mat[14] * m.mat[3];
		p[3] = mat[3] * m.mat[0] + mat[7] * m.mat[1] + mat[11] * m.mat[2] + mat[15] * m.mat[3];
		
		p[4] = mat[0] * m.mat[4] + mat[4] * m.mat[5] + mat[8] * m.mat[6] + mat[12] * m.mat[7];
		p[5] = mat[1] * m.mat[4] + mat[5] * m.mat[5] + mat[9] * m.mat[6] + mat[13] * m.mat[7];
		p[6] = mat[2] * m.mat[4] + mat[6] * m.mat[5] + mat[10] * m.mat[6] + mat[14] * m.mat[7];
		p[7] = mat[3] * m.mat[4] + mat[7] * m.mat[5] + mat[11] * m.mat[6] + mat[15] * m.mat[7];
		
		p[8] = mat[0] * m.mat[8] + mat[4] * m.mat[9] + mat[8] * m.mat[10] + mat[12] * m.mat[11];
		p[9] = mat[1] * m.mat[8] + mat[5] * m.mat[9] + mat[9] * m.mat[10] + mat[13] * m.mat[11];
		p[10] = mat[2] * m.mat[8] + mat[6] * m.mat[9] + mat[10] * m.mat[10] + mat[14] * m.mat[11];
		p[11] = mat[3] * m.mat[8] + mat[7] * m.mat[9] + mat[11] * m.mat[10] + mat[15] * m.mat[11];
		
		p[12] = mat[0] * m.mat[12] + mat[4] * m.mat[13] + mat[8] * m.mat[14] + mat[12] * m.mat[15];
		p[13] = mat[1] * m.mat[12] + mat[5] * m.mat[13] + mat[9] * m.mat[14] + mat[13] * m.mat[15];
		p[14] = mat[2] * m.mat[12] + mat[6] * m.mat[13] + mat[10] * m.mat[14] + mat[14] * m.mat[15];
		p[15] = mat[3] * m.mat[12] + mat[7] * m.mat[13] + mat[11] * m.mat[14] + mat[15] * m.mat[15];
		//@formatter:on

		System.arraycopy(p, 0, mat, 0, 16);
		identity = false;
		return this;
	}

	public Matrix44 scale(float x, float y, float z) {
		for (int i = 0; i < 4; i++) {
			mat[i] *= x;
			mat[4 + i] *= y;
			mat[8 + i] *= z;
		}
		return this;
	}

	public Matrix44 sub(Matrix44 m) {
		for (int i = 0; i < 16; i++) {
			mat[i] -= m.mat[i];
		}
		identity = false;
		return this;
	}

	public Matrix44 translate(float x, float y, float z) {
		Matrix44 tmp = new Matrix44();
		tmp.loadTranslate(x, y, z);
		return mult(tmp);
	}

}
