package org.droid3d.renderer;

import java.util.HashMap;
import java.util.List;
import java.util.Map;

import org.droid3d.math.Matrix44;
import org.droid3d.renderer.Material.BlendMode;
import org.droid3d.renderer.buffer.IndexBuffer;
import org.droid3d.renderer.buffer.VertexBuffer;
import org.droid3d.renderer.shader.Shader;
import org.droid3d.renderer.shader.ShaderProgram;
import org.droid3d.renderer.states.RenderState;
import org.droid3d.renderer.states.StateAlpha;
import org.droid3d.renderer.states.StateCulling;
import org.droid3d.scenegraph.Camera;
import org.droid3d.scenegraph.Culler;
import org.droid3d.scenegraph.Geometry;
import org.droid3d.scenegraph.GroupNode;
import org.droid3d.scenegraph.Scene;

/**
 *
 */
public abstract class AbstractRenderer implements Renderer, Pipeline {
	private final Map<String, Object>					mValues			= new HashMap<String, Object>();
	private final Map<RenderState.Type, RenderState>	mStates			= new HashMap<RenderState.Type, RenderState>();

	private final Map<VertexBuffer, Integer>			mVertexBuffers	= new HashMap<VertexBuffer, Integer>();
	private final Map<IndexBuffer, Integer>				mIndexBuffers	= new HashMap<IndexBuffer, Integer>();
	private final Map<Texture, Integer>					mTextures		= new HashMap<Texture, Integer>();
	private final Map<Shader, Integer>					mShaders		= new HashMap<Shader, Integer>();
	private final Map<ShaderProgram, Integer>			mShaderPrograms	= new HashMap<ShaderProgram, Integer>();

	protected ShaderProgram								mShaderProgram;
	protected VertexBuffer								mVertexBuffer;
	protected IndexBuffer								mIndexBuffer;
	protected Material									mMaterial;
	protected Texture									mTexture;

	protected Matrix44									mProjMatrix		= new Matrix44();
	protected Matrix44									mViewMatrix		= new Matrix44();
	protected Matrix44									mModelMatrix	= new Matrix44();

	protected Color										mClearColor		= Color.BLACK;
	protected RenderStatistics							mStatistics		= new RenderStatistics();

	public AbstractRenderer() {
		for (RenderState state : RenderState.DEFAULT) {
			mStates.put(state.getType(), state);
		}
	}

	@Override
	public final void delete(IndexBuffer buffer) {
		Integer handle = mIndexBuffers.get(buffer);
		if (handle != null) {
			deleteBuffer(handle);
			mIndexBuffers.remove(buffer);
			mStatistics.iboCount--;
			mStatistics.iboSize -= buffer.getBufferSize();
		}
	}

	@Override
	public final void delete(Shader shader) {
		Integer handle = mShaders.get(shader);
		if (handle != null) {
			deleteShader(handle);
			mShaders.remove(shader);
			mStatistics.shaderCount--;
		}
	}

	@Override
	public final void delete(ShaderProgram program) {
		Integer handle = mShaderPrograms.get(program);
		if (handle != null) {
			deleteShader(handle);
			mShaderPrograms.remove(program);
			mStatistics.shaderProgramCount--;
		}
	}

	@Override
	public final void delete(Texture texture) {
		Integer handle = mTextures.get(texture);
		if (handle != null) {
			deleteTexture(handle);
			mTextures.remove(texture);
			mStatistics.textureCount--;
			mStatistics.textureSize -= texture.getBitmap().getByteCount();
		}
	}

	@Override
	public final void delete(VertexBuffer buffer) {
		Integer handle = mVertexBuffers.get(buffer);
		if (handle != null) {
			deleteBuffer(handle);
			mVertexBuffers.remove(buffer);
			mStatistics.vboCount--;
			mStatistics.vboSize -= buffer.getBufferSize();
		}

	}

	@Override
	public final void deleteAll() {
		for (Integer handle : mIndexBuffers.values()) {
			deleteBuffer(handle);
		}
		mIndexBuffers.clear();
		mStatistics.iboCount = 0;
		mStatistics.iboSize = 0;

		for (Integer handle : mVertexBuffers.values()) {
			deleteBuffer(handle);
		}
		mVertexBuffers.clear();
		mStatistics.vboCount = 0;
		mStatistics.vboSize = 0;

		for (Integer handle : mTextures.values()) {
			deleteTexture(handle.intValue());
		}
		mTextures.clear();
		mStatistics.textureCount = 0;
		mStatistics.textureSize = 0;

		for (Integer handle : mShaderPrograms.values()) {
			deleteShaderProgram(handle);
		}
		mShaderPrograms.clear();
		mStatistics.shaderProgramCount = 0;

		for (Integer handle : mShaders.values()) {
			deleteShader(handle);
		}
		mShaders.clear();
		mStatistics.shaderCount = 0;
	}

	@Override
	public final void draw(List<Geometry> objects) {
		for (Geometry obj : objects) {
			draw(obj);
		}

	}

	@Override
	public final void draw(Scene scene) {
		GroupNode root = scene.getRoot();
		Camera camera = scene.getCamera();
		Culler culler = scene.getCuller();
		culler.setCamera(camera);
		culler.computeVisible(root);

		setCamera(camera);
		draw(culler.getVisible());
	}

	@Override
	public void drawElements(int mode, int count, int offset) {
		mStatistics.drawCalls++;
	}

	@Override
	public final void enable(IndexBuffer buffer) {
		Integer handle = mIndexBuffers.get(buffer);
		if (handle == null) {
			handle = upload(buffer);
		}
		bindIndexBuffer(handle);
		mIndexBuffer = buffer;
	}

	@Override
	public final void enable(List<Light> lights) {
		int lightCount = lights.size();
		if (lightCount > 0) {
			for (int i = 0; i < lightCount; i++) {
				setLight(lights.get(i), i);
			}
			enableLighting();
		}
	}

	@Override
	public final void enable(ShaderProgram shaderProgram) {
		Integer programHandle = mShaderPrograms.get(shaderProgram);
		if (programHandle == null) {
			Shader shader = shaderProgram.getVertexShader();
			Integer vertShaderHandle = mShaders.get(shader);
			if (vertShaderHandle == null) {
				vertShaderHandle = compileShader(shader);
				if (vertShaderHandle > 0) {
					mShaders.put(shader, vertShaderHandle);
				}
			}

			shader = shaderProgram.getFragmentShader();
			Integer fragShaderHandle = mShaders.get(shader);
			if (fragShaderHandle == null) {
				fragShaderHandle = compileShader(shader);
				if (fragShaderHandle > 0) {
					mShaders.put(shader, fragShaderHandle);
				}
			}

			programHandle = linkShaderProgram(vertShaderHandle,
					fragShaderHandle);
		}
		enableShaderProgram(programHandle);
	}

	@Override
	public final void enable(Texture texture, BlendMode mode, int unit) {
		Integer handle = mTextures.get(texture);
		if (handle == null) {
			handle = upload(texture);
		}
		bindTexture(handle, mode, unit);
	}

	@Override
	public void enable(VertexBuffer buffer) {
		Integer handle = mVertexBuffers.get(buffer);
		if (handle == null) {
			handle = upload(buffer);
		}
		bindVertexBuffer(handle);
	}

	@SuppressWarnings("unchecked")
	@Override
	public final <V> V get(String key) {
		return (V) mValues.get(key);
	}

	@Override
	public final Matrix44 getProjMatrix() {
		return mProjMatrix;
	}

	@Override
	public final RenderStatistics getStatistics() {
		return mStatistics;
	}

	@Override
	public final Matrix44 getViewMatrix() {
		return mViewMatrix;
	}

	@Override
	public void onInitialize() {
		for (RenderState state : RenderState.DEFAULT) {
			mStates.put(state.getType(), state);
		}
	}

	@Override
	public void onPreDraw() {
		mStatistics.drawCalls = 0;
		mStatistics.frameNumber++;
	}

	protected abstract void onProjMatrixChanged(Matrix44 proj);

	protected abstract void onViewMatrixChanged(Matrix44 view);

	@Override
	public final void reloadAll() {
		for (RenderState state : mStates.values()) {
			enable(state);
		}

		for (VertexBuffer buffer : mVertexBuffers.keySet()) {
			int handle = uploadVertexBuffer(buffer);
			if (handle > 0) {
				mVertexBuffers.put(buffer, handle);
			}
		}

		for (IndexBuffer buffer : mIndexBuffers.keySet()) {
			int handle = uploadIndexBuffer(buffer);
			if (handle > 0) {
				mIndexBuffers.put(buffer, handle);
			}
		}

		for (Texture texture : mTextures.keySet()) {
			int handle = uploadTexture(texture);
			if (handle > 0) {
				mTextures.put(texture, handle);
			}
		}

		// for (ShaderProgram shader : mShaderPrograms.keySet()) {
		// int handle = linkShaderProgram(shader);
		// if (handle > 0) {
		// mShaderPrograms.put(shader, handle);
		// }
		// }
	}

	@Override
	public final void set(Map<RenderState.Type, RenderState> states) {
		RenderState state = states.get(RenderState.Type.Alpha);
		if (state != null && state != mStates.get(RenderState.Type.Alpha)) {
			setState((StateAlpha) state);
		}

		state = states.get(RenderState.Type.Culling);
		if (state != null && state != mStates.get(RenderState.Type.Culling)) {
			setState((StateCulling) state);
		}

		state = states.get(RenderState.Type.DepthBuffer);
		if (state != null && state != mStates.get(RenderState.Type.DepthBuffer)) {
			setState((StateCulling) state);
		}
	}

	@Override
	public final void set(String key, Object value) {
		mValues.put(key, value);
	}

	@Override
	public final void setCamera(Camera camera) {
		setViewPort(camera.getViewportX(), camera.getViewPortY(),
				camera.getViewportWidth(), camera.getViewportHeight());
		setProjMatrix(camera.getProjectionMatrix(mProjMatrix));
		setViewMatrix(camera.getViewMatrix(mViewMatrix));
	}

	@Override
	public final void setClearColor(Color color) {
		mClearColor = color;
	}

	@Override
	public final void setProjMatrix(Matrix44 proj) {
		mProjMatrix.load(proj);
		onProjMatrixChanged(mProjMatrix);
	}

	private void setState(StateAlpha state) {
		// TODO Auto-generated method stub
	}

	private void setState(StateCulling state) {
		// TODO Auto-generated method stub

	}

	@Override
	public final void setViewMatrix(Matrix44 view) {
		mViewMatrix.load(view);
		onViewMatrixChanged(mViewMatrix);
	}

	@Override
	public final void update(IndexBuffer buffer) {
		Integer handle = mIndexBuffers.get(buffer);
		if (handle != null) {
			updateIndexBuffer(buffer, handle);
		}
	}

	@Override
	public final void update(VertexBuffer buffer) {
		Integer handle = mVertexBuffers.get(buffer);
		if (handle != null) {
			updateVertexBuffer(buffer, handle);
		}
	}

	@Override
	public final int upload(IndexBuffer buffer) {
		int handle = uploadIndexBuffer(buffer);
		if (handle > 0) {
			mIndexBuffers.put(buffer, handle);
			mStatistics.iboCount++;
			mStatistics.iboSize += buffer.getBufferSize();
		}
		return handle;
	}

	@Override
	public final int upload(Shader shader) {
		int handle = compileShader(shader);
		if (handle > 0) {
			mShaders.put(shader, handle);
			mStatistics.shaderCount++;
		}
		return handle;
	}

	@Override
	public final int upload(ShaderProgram shaderProgram) {
		// List<Shader> shader = shaderProgram.getShader();
		// int[] shaderHandles = new int[shader.size()];
		// for (int i = 0; i < shader.size(); i++) {
		// Shader s = shader.get(i);
		// if (mShaders.get(s) == null) {
		// shaderHandles[i] = upload(s);
		// }
		// }
		//
		// int handle = linkShaderProgram(shaderHandles);
		// if (handle > 0) {
		// mShaderPrograms.put(shaderProgram, handle);
		// mStatistics.shaderProgramCount++;
		// }
		return 0;
	}

	@Override
	public final int upload(Texture texture) {
		int handle = uploadTexture(texture);
		if (handle > 0) {
			mTextures.put(texture, handle);
			mStatistics.textureCount++;
			mStatistics.textureSize += texture.getSize();
		}
		return handle;
	}

	@Override
	public final int upload(VertexBuffer buffer) {
		int handle = uploadVertexBuffer(buffer);
		if (handle > 0) {
			mVertexBuffers.put(buffer, handle);
			mStatistics.vboCount++;
			mStatistics.vboSize += buffer.getBufferSize();
		}
		return handle;
	}
}
