/*******************************************************************************
 * Copyright 2013 Bo Meier (bo.meier@gmail.com)
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 *   http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 ******************************************************************************/
package org.droid3d.renderer;

import java.util.ArrayList;
import java.util.List;

import org.droid3d.renderer.shader.ShaderProgram;

/**
 * Defines the properties of a material.
 */
public class Material {
	public enum BlendMode {
		REPLACE, MODULATE, DECAL;
	}

	/**
	 * Holds a {@link Texture} and a {@link BlendMode}.
	 */
	private static class TextureBlendMode {
		Texture		texture;
		BlendMode	blendMode;
	}

	// defaults
	public static Color				DEFAULT_AMBIENT_COLOR	= new Color(0.2f, 0.2f, 0.2f);
	public static Color				DEFAULT_DIFFUSE_COLOR	= new Color(0.8f, 0.8f, 0.8f);
	public static Color				DEFAULT_SPECULAR_COLOR	= new Color(1.0f, 1.0f, 1.0f);

	public static float				DEFAULT_SHININESS		= 0.0f;
	public static float				DEFAULT_TRANSPARENCY	= 1.0f;

	private String					mName;
	private ShaderProgram			mShaderProgram;

	// colors
	private Color					mAmbientColor			= DEFAULT_AMBIENT_COLOR;
	private Color					mDiffuseColor			= DEFAULT_DIFFUSE_COLOR;
	private Color					mSpecular				= DEFAULT_SPECULAR_COLOR;

	private float					mShininess				= DEFAULT_SHININESS;
	private float					mTranspanrency			= DEFAULT_TRANSPARENCY;

	private List<TextureBlendMode>	mTextures				= new ArrayList<TextureBlendMode>();

	public void addTexture(Texture texture, BlendMode mode) {
		TextureBlendMode tbm = new TextureBlendMode();
		tbm.texture = texture;
		tbm.blendMode = mode;
		mTextures.add(tbm);
	}

	public Color getAmbientColor() {
		return mAmbientColor;
	}

	public BlendMode getBlendMode(int texture) {
		return mTextures.get(texture).blendMode;
	}

	public Color getDiffuseColor() {
		return mDiffuseColor;
	}

	public String getName() {
		return mName;
	}

	public int getNumTexture() {
		return mTextures.size();
	}

	public ShaderProgram getShaderProgram() {
		return mShaderProgram;
	}

	public float getShininess() {
		return mShininess;
	}

	public Color getSpecularColor() {
		return mSpecular;
	}

	public Texture getTexture(int texture) {
		return mTextures.get(texture).texture;
	}

	public float getTranspanrency() {
		return mTranspanrency;
	}

	public void setAmbientColor(Color ambientColor) {
		mAmbientColor = ambientColor;
	}

	public void setColor(Color color) {
		setAmbientColor(color);
		setDiffuseColor(color);
	}

	public void setDiffuseColor(Color color) {
		mDiffuseColor = color;

	}

	public void setName(String name) {
		mName = name;
	}

	public void setShaderProgram(ShaderProgram shaderProgram) {
		mShaderProgram = shaderProgram;
	}

	public void setShininess(float shininess) {
		mShininess = shininess;
	}

	public void setSpecularColor(Color color) {
		mSpecular = color;
	}

	public void setTranspanrency(float transpanrency) {
		mTranspanrency = transpanrency;
	}
}
