/*******************************************************************************
 * Copyright 2013 Bo Meier (bo.meier@gmail.com)
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 *   http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 ******************************************************************************/
package org.droid3d.renderer;

import org.droid3d.renderer.buffer.IndexBuffer;
import org.droid3d.renderer.buffer.VertexBuffer;
import org.droid3d.scenegraph.BoundingVolume;
import org.droid3d.utils.BoundingVolumes;

/**
 * A mesh of vertices.
 * <p/>
 * The mesh is created of the vertices of a {@link VertexBuffer} that are
 * pointed to by indices of an {@link IndexBuffer}.
 * 
 * 
 */
public class Mesh {

	private final VertexBuffer	mVertexBuffer;

	/**
	 * offset of the first vertex
	 */
	private int					mStartVertex;
	private int					mVertexCount;

	private final IndexBuffer	mIndexBuffer;
	private int					mStartIndex;
	private int					mIndexCount;

	/**
	 * the dirty flag is set to true whenever mStartIndex, mStartVertex,
	 * mIndexCount or mVertexCounts changes.
	 */
	private boolean				mDirty;

	/**
	 * the bounding volume of the mesh
	 */
	private BoundingVolume		mBoundingVolume;

	private RenderMode			mRenderMode;

	/**
	 * Create a mesh from a VertexBuffer and an IndexBuffer.
	 * 
	 * @param vertexBuffer
	 * @param indexBuffer
	 */
	public Mesh(VertexBuffer vertexBuffer, IndexBuffer indexBuffer) {
		this(vertexBuffer, indexBuffer, 0, indexBuffer.getIndexCount(), RenderMode.TRIANGLES);
	}

	public Mesh(VertexBuffer vertexBuffer, IndexBuffer indexBuffer, int startIndex, int indexCount) {
		this(vertexBuffer, indexBuffer, startIndex, indexCount, RenderMode.TRIANGLES);
	}

	public Mesh(VertexBuffer vertexBuffer, IndexBuffer indexBuffer, int startIndex, int indexCount,
			RenderMode renderMode) {
		mVertexBuffer = vertexBuffer;
		mIndexBuffer = indexBuffer;
		mStartIndex = startIndex;
		mIndexCount = indexCount;
		mRenderMode = renderMode;
	}

	/**
	 * @param vertexBuffer
	 * @param indexBuffer
	 */
	public Mesh(VertexBuffer vertexBuffer, IndexBuffer indexBuffer, RenderMode renderMode) {
		this(vertexBuffer, indexBuffer, 0, indexBuffer.getIndexCount(), renderMode);
	}

	/**
	 * Get the index buffer.
	 * 
	 * @return
	 */
	public IndexBuffer getIndexBuffer() {
		return mIndexBuffer;
	}

	/**
	 * Get the number of indices.
	 * 
	 * @return
	 */
	public int getIndexCount() {
		return mIndexCount;
	}

	/**
	 * Get the render mode.
	 * 
	 * @return
	 */
	public RenderMode getRenderMode() {
		return mRenderMode;
	}

	public int getStartIndex() {
		return mStartIndex;
	}

	public int getStartVertex() {
		return mStartVertex;
	}

	/**
	 * Get the vertex buffer.
	 * 
	 * @return
	 */
	public VertexBuffer getVertexBuffer() {
		return mVertexBuffer;
	}

	public int getVertexCount() {
		return mVertexCount;
	}

	/**
	 * Get the dirty flag. The mesh's dirty flag is set to true whenever the
	 * index count or start index is modified.
	 * 
	 * @return
	 */
	public boolean isDirty() {
		return mDirty;
	}

	/**
	 * Will not update models bounding volume!
	 * 
	 * @param indexCount
	 */
	public void setIndexCount(int indexCount) {
		setIndexCount(indexCount, false);
	}

	public void setIndexCount(int indexCount, boolean updateModelBound) {
		if (mIndexCount != indexCount) {
			mIndexCount = indexCount;
			mDirty = true;
			if (updateModelBound) {
				updateBoundingVolume();
			}
		}
	}

	/**
	 * Will not update models bounding volume!
	 * 
	 * @param startIndex
	 */
	public void setStartIndex(int startIndex) {
		setStartIndex(startIndex, false);
	}

	public void setStartIndex(int startIndex, boolean updateModelBound) {
		if (mStartIndex != startIndex) {
			mStartIndex = startIndex;
			mDirty = true;
			if (updateModelBound) {
				updateBoundingVolume();
			}
		}
	}

	/**
	 * Update the bounding volume of the mesh.
	 */
	public void updateBoundingVolume() {
		if (mStartIndex == 0 && mIndexCount == mIndexBuffer.getIndexCount()) {
			mBoundingVolume.load(BoundingVolumes.computeFrom(mVertexBuffer));
		} else {
			mBoundingVolume.load(mVertexBuffer, mIndexBuffer, mStartIndex, mIndexCount);
		}
	}
}
