/*******************************************************************************
 * Copyright 2013 Bo Meier (bo.meier@gmail.com)
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 *   http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 ******************************************************************************/
package org.droid3d.renderer;

import java.util.List;
import java.util.Map;

import org.droid3d.math.Matrix44;
import org.droid3d.renderer.Material.BlendMode;
import org.droid3d.renderer.buffer.IndexBuffer;
import org.droid3d.renderer.buffer.VertexBuffer;
import org.droid3d.renderer.shader.Shader;
import org.droid3d.renderer.shader.ShaderProgram;
import org.droid3d.renderer.states.RenderState;
import org.droid3d.scenegraph.Camera;
import org.droid3d.scenegraph.Geometry;
import org.droid3d.scenegraph.Scene;

public interface Renderer {

	/**
	 * Delete the data of an IndexBuffer from video memory.
	 * 
	 * @param buffer
	 */
	public void delete(IndexBuffer buffer);

	public void delete(Shader shader);

	public void delete(ShaderProgram shaderProgram);

	/**
	 * Delete a Texture from video memory.
	 * 
	 * @param texture
	 */
	public void delete(Texture texture);

	/**
	 * Delete the data of a VertexBuffer from video memory.
	 * 
	 * @param buffer
	 */
	public void delete(VertexBuffer buffer);

	/**
	 * Delete all resources from video memory.
	 */
	public void deleteAll();

	public void disableLighting();

	/**
	 * Draw an object.
	 */
	public void draw(Geometry geometry);

	public void draw(List<Geometry> geometeries);

	/**
	 * Draw a mesh.
	 */
	public void draw(Mesh mesh);

	public void draw(Scene scene);

	public void drawElements(int mode, int count, int offset);

	/**
	 * @param buffer
	 */
	public void enable(IndexBuffer buffer);

	public void enable(List<Light> lights);

	public void enable(Material material);

	public void enable(RenderState state);

	public void enable(ShaderProgram shaderProgram);

	public void enable(Texture texture, BlendMode mode, int unit);

	public void enable(VertexBuffer buffer);

	public void enableLighting();

	public <V> V get(String key);

	public Matrix44 getProjMatrix();

	public RenderStatistics getStatistics();

	public Matrix44 getViewMatrix();

	public void onInitialize();

	public void onPostDraw();

	public void onPreDraw();

	public void reloadAll();

	public void set(Map<RenderState.Type, RenderState> states);

	public void set(String key, Object value);

	public void setCamera(Camera camera);

	public void setClearColor(Color color);

	public void setLight(Light light, int index);

	public void setProjMatrix(Matrix44 proj);

	public void setViewMatrix(Matrix44 view);

	public void setViewPort(int x, int y, int width, int height);

	public void update(IndexBuffer buffer);

	/**
	 * @param buffer
	 */
	public void update(VertexBuffer buffer);

	/**
	 * Upload the data of an IndexBuffer to a buffer in video memory.
	 * 
	 * @return handle
	 */
	public int upload(IndexBuffer buffer);

	public int upload(Shader shader);

	public int upload(ShaderProgram shaderProgram);

	/**
	 * Upload a Texture2D to video memory.
	 * 
	 * @param texture
	 * @return
	 */
	public int upload(Texture texture);

	/**
	 * Upload the data of a VertexBuffer to a buffer in video memory.
	 * 
	 * @return handle
	 */
	public int upload(VertexBuffer buffer);
}
