/*******************************************************************************
 * Copyright 2013 Bo Meier (bo.meier@gmail.com)
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 *   http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 ******************************************************************************/
package org.droid3d.renderer.buffer;

import java.nio.ByteBuffer;
import java.nio.ByteOrder;

/**
 *
 */
public class Buffer {
	public enum Type {
		VERTEX_BUFFER, INDEX_BUFFER;
	}

	public enum Usage {
		STATIC, DYNAMIC, STREAM;
	}

	private ByteBuffer	mData;
	private final Usage	mUsage;
	private final Type	mType;

	public Buffer(int size, Buffer.Type type) {
		this(size, type, Usage.STATIC);
	}

	public Buffer(int size, Buffer.Type type, Usage usage) {
		mUsage = usage;
		mType = type;
		allocateBuffer(size);
	}

	/**
	 * Allocate the buffer.
	 * 
	 * @param size
	 *            size in bytes
	 */
	protected void allocateBuffer(int size) {
		mData = ByteBuffer.allocateDirect(size).order(ByteOrder.nativeOrder());
		mData.clear();
	}

	/**
	 * Get the buffer size.
	 * 
	 * @return size in bytes.
	 */
	public int getBufferSize() {
		return mData.capacity();
	}

	public byte getByte(int position) {
		mData.position(0);
		return mData.get();
	}

	public byte[] getByte(int position, byte[] b) {
		mData.position(0);
		mData.get(b);
		return b;
	}

	public byte[] getByte(int position, int count) {
		byte[] b = new byte[count];
		return getByte(position, b);
	}

	public ByteBuffer getData() {
		return mData;
	}

	public float getFloat(int position) {
		mData.position(position);
		return mData.asFloatBuffer().get();
	}

	public float[] getFloat(int position, float[] array) {
		mData.position(position);
		mData.asFloatBuffer().get(array);
		return array;
	}

	public float[] getFloat(int position, int length) {
		float[] array = new float[length];
		return getFloat(position, array);
	}

	public int getInt(int position) {
		mData.position(position);
		return mData.asIntBuffer().get();
	}

	public int[] getInt(int position, int length) {
		int[] tmp = new int[length];
		return getInt(position, tmp);
	}

	public int[] getInt(int position, int[] array) {
		mData.position(position);
		mData.asIntBuffer().get(array);
		return array;
	}

	public short getShort(int position) {
		mData.position(position);
		return mData.asShortBuffer().get();
	}

	public short[] getShort(int position, int length) {
		short[] tmp = new short[length];
		return getShort(position, tmp);
	}

	public short[] getShort(int position, short[] array) {
		mData.position(position);
		mData.asShortBuffer().get(array);
		return array;
	}

	public Type getType() {
		return mType;
	}

	public Usage getUsage() {
		return mUsage;
	}

	public void putByte(int position, byte... b) {
		mData.position(position);
		mData.put(b);
	}

	public void putFloat(int position, float... v) {
		mData.position(position);
		mData.asFloatBuffer().put(v);
	}

	public void putInt(int position, int... i) {
		mData.position(position);
		mData.asIntBuffer().put(i);
	}

	public void putShort(int position, short... s) {
		mData.position(position);
		mData.asShortBuffer().put(s);
	}
}
