/*******************************************************************************
 * Copyright 2013 Bo Meier (bo.meier@gmail.com)
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 *   http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 ******************************************************************************/
package org.droid3d.renderer.buffer;

import org.droid3d.utils.Type;

/**
 * Definition of a vertex attribute.
 * <p/>
 * A vertex attribute is defined by its name, type and the number of elements of
 * the attributes type.
 */
public class VertexAttribute {
	/**
	 *
	 */
	public enum Usage {
		Position, Normal, Color, TexCoord, Generic;
	}

	private String		mName;
	private final int	mElementCount;
	private final Type	mType;
	private final int	mSize;
	private final Usage	mUsage;

	/**
	 * Create a VertexAttribute.
	 * 
	 * @param name
	 * @param type
	 * @param elementCount
	 */
	public VertexAttribute(String name, Type type, int elementCount) {
		this(name, type, elementCount, Usage.Generic);
	}

	/**
	 * 
	 * @param name
	 *            name of the attribute
	 * @param type
	 *            type of the attribute
	 * @param elementCount
	 *            number of elements of given type
	 * @param usage
	 */
	public VertexAttribute(String name, Type type, int elementCount, Usage usage) {
		mName = name;
		mElementCount = elementCount;
		mType = type;
		mSize = mElementCount * mType.size();
		mUsage = usage;
	}

	/**
	 * Get the number of elements.
	 * 
	 * @return
	 */
	public int getElementCount() {
		return mElementCount;
	}

	/**
	 * Get the name.
	 * 
	 * @return name of the attribute
	 */
	public String getName() {
		return mName;
	}

	/**
	 * Get the size of the attribute.
	 * 
	 * @return size in bytes
	 */
	public int getSize() {
		return mSize;
	}

	/**
	 * Get the type of the vertex attribute.
	 * 
	 * @return type of the attribute
	 */
	public Type getType() {
		return mType;
	}

	public Usage getUsage() {
		return mUsage;
	}

	/**
	 * Set the name of the vertex attribute.
	 * 
	 * @param name
	 *            the name of the vertex attribute
	 */
	public void setName(String name) {
		mName = name;
	}
}
