/*******************************************************************************
 * Copyright 2013 Bo Meier (bo.meier@gmail.com)
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 *   http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 ******************************************************************************/
package org.droid3d.renderer.buffer;

import java.util.Arrays;
import java.util.Collection;

/**
 * A buffer to store vertices.
 * 
 */
public class VertexBuffer extends Buffer {
	public interface Chan {

	}

	public interface Factory {
		public VertexBuffer create(VertexDescriptor format, int vertexCount);
	}

	protected int				mVertexCount;
	protected VertexDescriptor	mVertexDescriptor;

	/**
	 * Create a VertexBuffer.
	 * 
	 * @param vertexCount
	 *            the number of vertices stored in the buffer
	 * @param vertexAttributes
	 *            the attributes of the vertices stored in the buffer
	 */
	public VertexBuffer(int vertexCount, Collection<VertexAttribute> vertexAttributes) {
		this(vertexCount, Usage.STATIC, new VertexDescriptor(vertexAttributes));
	}

	/**
	 * Create a VertexBuffer.
	 * 
	 * @param vertexCount
	 *            the number of vertices stored in the buffer
	 * @param vertexAttributes
	 *            the attributes of the vertices stored in the buffer
	 */
	public VertexBuffer(int vertexCount, Usage usage, Collection<VertexAttribute> vertexAttributes) {
		this(vertexCount, usage, new VertexDescriptor(vertexAttributes));
	}

	/**
	 * Create a VertexBuffer.
	 * 
	 * @param vertexCount
	 *            the number of vertices stored in the buffer
	 * @param vertexAttributes
	 *            the attributes of the vertices stored in the buffer
	 */
	public VertexBuffer(int vertexCount, Usage usage, VertexAttribute... vertexAttributes) {
		this(vertexCount, usage, Arrays.asList(vertexAttributes));
	}

	/**
	 * Creates a VertexBuffer.
	 * 
	 * @param vertexCount
	 *            the number of vertices stored in the buffer
	 * @param vertexDescriptor
	 *            the definition of the vertices stored in the buffer
	 */
	public VertexBuffer(int vertexCount, Usage usage, VertexDescriptor vertexDescriptor) {
		super(vertexCount * vertexDescriptor.getVertexSize(), Buffer.Type.VERTEX_BUFFER, usage);
		mVertexDescriptor = vertexDescriptor;
		mVertexCount = vertexCount;
	}

	/**
	 * Create a VertexBuffer.
	 * 
	 * @param vertexCount
	 *            the number of vertices stored in the buffer
	 * @param vertexAttributes
	 *            the attributes of the vertices stored in the buffer
	 */
	public VertexBuffer(int vertexCount, VertexAttribute... vertexAttributes) {
		this(vertexCount, Usage.STATIC, Arrays.asList(vertexAttributes));
	}

	/**
	 * Creates a VertexBuffer.
	 * 
	 * @param vertexCount
	 *            the number of vertices stored in the buffer
	 * @param vertexDescriptor
	 *            the definition of the vertices stored in the buffer
	 */
	public VertexBuffer(int vertexCount, VertexDescriptor vertexDescriptor) {
		this(vertexCount, Usage.STATIC, vertexDescriptor);
	}

	/**
	 * Get the definition of a vertex attribute.
	 * 
	 * @param i
	 *            the index of the attribute
	 * @return
	 */
	public VertexAttribute getAttribute(int i) {
		return mVertexDescriptor.getAttribute(i);
	}

	/**
	 * Get the definition of a vertex attribute.
	 * 
	 * @param name
	 *            the name of the attribute
	 * @return
	 */
	public VertexAttribute getAttribute(String name) {
		return mVertexDescriptor.getAttribute(name);
	}

	public VertexAttribute getAttribute(VertexAttribute.Usage usage) {
		return getVertexDescriptor().getAttribute(usage);
	}

	/**
	 * Get the number of vertex attributes.
	 * 
	 * @return
	 */
	public int getAttributeCount() {
		return mVertexDescriptor.getAttributeCount();
	}

	/**
	 * Get the offset of a vertex attribute.
	 * 
	 * @param i
	 *            the index of the attribute
	 * @return offset in bytes
	 */
	public int getOffset(int i) {
		return mVertexDescriptor.getOffset(i);
	}

	/**
	 * Get the offset of a vertex attribute.
	 * 
	 * @param attr
	 * @return offset in bytes
	 */
	public int getOffset(VertexAttribute attr) {
		return mVertexDescriptor.getOffset(attr);
	}

	/**
	 * Returns the number of vertices in this buffer.
	 * 
	 * @return number of vertices
	 */
	public int getVertexCount() {
		return mVertexCount;
	}

	/**
	 * Returns the definition of the vertices stored in the buffer.
	 * 
	 * @return the definition of the vertices stored in the buffer
	 */
	public VertexDescriptor getVertexDescriptor() {
		return mVertexDescriptor;
	}

	/**
	 * Get the size of a vertex.
	 * 
	 * @return size in bytes
	 */
	public int getVertexSize() {
		return mVertexDescriptor.getVertexSize();
	}

	public boolean hasAttribute(String name) {
		return mVertexDescriptor.hasAttribute(name);
	}

	public boolean hasAttribute(VertexAttribute positionAttr) {
		return mVertexDescriptor.hasAttribute(positionAttr);
	}

	public boolean hasAttribute(VertexAttribute.Usage usage) {
		return mVertexDescriptor.hasAttribute(usage);
	}
}
