/*******************************************************************************
 * Copyright 2013 Bo Meier (bo.meier@gmail.com)
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 *   http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 ******************************************************************************/
package org.droid3d.renderer.buffer;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

/**
 * A VertexDescriptor defines the attributes of a vertex.
 */
public class VertexDescriptor {

	private List<VertexAttribute>			mAttributes	= new ArrayList<VertexAttribute>();
	private Map<VertexAttribute, Integer>	mOffsets	= new HashMap<VertexAttribute, Integer>();
	private int								mVertexSize	= 0;

	/**
	 * @param attributes
	 */
	public VertexDescriptor(Collection<VertexAttribute> attributes)
	{
		mAttributes.addAll(attributes);
		updateOffsetsAndSize();
	}

	/**
	 * @param attributes
	 */
	public VertexDescriptor(VertexAttribute... attributes)
	{
		this(Arrays.asList(attributes));
	}

	/**
	 * Get the definition of a vertex attribute.
	 * 
	 * @param i
	 *            the index of the attribute
	 * @return the definition of the attribute
	 */
	public VertexAttribute getAttribute(int i)
	{
		return mAttributes.get(i);
	}

	/**
	 * Get the definition of a vertex attribute.
	 * 
	 * @param name
	 *            the name of the attribute
	 * @return the definition of the attribute, if a definition for an attribute
	 *         with the given name exists, null otherwise
	 */
	public VertexAttribute getAttribute(String name)
	{
		for (VertexAttribute attr : mAttributes) {
			if (name.equals(attr.getName())) {
				return attr;
			}
		}
		return null;
	}

	public VertexAttribute getAttribute(VertexAttribute.Usage usage)
	{
		List<VertexAttribute> attributes = getAttributes(usage);
		return attributes.size() > 0 ? attributes.get(0) : null;
	}

	/**
	 * Get the number of vertex attributes.
	 * 
	 * @return number of attributes
	 */
	public int getAttributeCount()
	{
		return mAttributes.size();
	}

	public List<VertexAttribute> getAttributes()
	{
		return new ArrayList<VertexAttribute>(mAttributes);
	}

	public List<VertexAttribute> getAttributes(VertexAttribute.Usage usage)
	{
		List<VertexAttribute> attributes = new ArrayList<VertexAttribute>();
		for (VertexAttribute attribute : mAttributes) {
			if (attribute.getUsage() == usage) {
				attributes.add(attribute);
			}
		}
		return attributes;
	}

	/**
	 * Get the byte offset of a vertex attribute.
	 * 
	 * @param i
	 *            the index of the attribute
	 * @return the offset of the attribute in bytes
	 */
	public int getOffset(int i)
	{
		VertexAttribute attr = mAttributes.get(i);
		return getOffset(attr);
	}

	/**
	 * Get the byte offset of a vertex attribute.
	 * 
	 * @param attr
	 *            the attribute
	 * @return the offset of the attribute in bytes
	 */
	public int getOffset(VertexAttribute attr)
	{
		return mOffsets.get(attr);
	}

	/**
	 * Get the size of a vertex in bytes.
	 * 
	 * @return the size of the vertex
	 */
	public int getVertexSize()
	{
		return mVertexSize;
	}

	/**
	 * Checks if the vertex descriptor defines a vertex attribute.
	 * 
	 * @param name
	 *            the name of the attribute
	 * @return
	 */
	public boolean hasAttribute(String name)
	{
		for (VertexAttribute attr : mAttributes) {
			if (name.equals(attr.getName())) {
				return true;
			}
		}
		return false;
	}

	/**
	 * Checks if the vertex descriptor defines a vertex attribute.
	 * 
	 * @param attribute
	 * @return
	 */
	public boolean hasAttribute(VertexAttribute attribute)
	{
		return mAttributes.contains(attribute);
	}

	public boolean hasAttribute(VertexAttribute.Usage usage)
	{
		for (VertexAttribute attr : mAttributes) {
			if (attr.getUsage() == usage) {
				return true;
			}
		}
		return false;
	}

	/**
	 * Calculates the offsets of the vertex attributes and the size of a vertex.
	 */
	private void updateOffsetsAndSize()
	{
		mOffsets.clear();
		mVertexSize = 0;

		if (mAttributes.size() > 0) {
			VertexAttribute attr = mAttributes.get(0);
			mVertexSize = attr.getSize();
			mOffsets.put(attr, Integer.valueOf(0));
			int offset = 0;
			for (int i = 1; i < mAttributes.size(); i++) {
				attr = mAttributes.get(i);
				offset += mAttributes.get(i - 1).getSize();
				mOffsets.put(attr, offset);
				mVertexSize += attr.getSize();
			}
		}
	}
}
