/*******************************************************************************
 * Copyright 2013 Bo Meier (bo.meier@gmail.com)
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 *   http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 ******************************************************************************/
package org.droid3d.scenegraph;

import org.droid3d.math.Float3;
import org.droid3d.renderer.buffer.IndexBuffer;
import org.droid3d.renderer.buffer.VertexBuffer;
import org.droid3d.utils.BoundingVolumes;
import org.droid3d.utils.buffer.Channel;
import org.droid3d.utils.buffer.ChannelFactory;

public class BoundingVolume {

	private Float3	mCenter	= new Float3();
	private float	mRadius	= 0f;

	public BoundingVolume() {
	}

	public BoundingVolume(BoundingVolume bv) {
		load(bv);
	}

	public BoundingVolume(Float3 center, float radius) {
		load(center, radius);
	}

	public Float3 getCenter() {
		return mCenter;
	}

	public float getRadius() {

		return mRadius;
	}

	public void load(BoundingVolume bv) {
		load(bv.mCenter, bv.mRadius);
	}

	public void load(Float3 center, float radius) {
		mCenter = center;
		mRadius = radius;
	}

	public void load(VertexBuffer vertexBuffer, IndexBuffer indexBuffer, int startIndex, int indexCount) {
		mCenter = new Float3();
		mRadius = 0f;

		Channel<Float3> positionAdapter = ChannelFactory.newPositionChannel(vertexBuffer);
		if (positionAdapter != null) {
			for (int i = 0; i < indexCount; i++) {
				short index = indexBuffer.getIndex(i + startIndex);
				mCenter.add(positionAdapter.read(index));
			}
			mCenter.mult(1f / indexCount);

			for (int i = 0; i < indexCount; i++) {
				short index = indexBuffer.getIndex(i + startIndex);
				Float3 diff = positionAdapter.read(index);
				diff.sub(mCenter);
				float rSqr = diff.getSquaredLength();
				if (rSqr > mRadius) {
					mRadius = rSqr;
				}
			}
			mRadius = (float) Math.sqrt(mRadius);
		}

	}

	public void merge(BoundingVolume bv) {
		Float3 centerDiff = Float3.sub(mCenter, bv.mCenter);
		float centerDiffSqrLength = centerDiff.getSquaredLength();
		float radiusDiff = mRadius - bv.mRadius;
		float radiusDiffSqr = radiusDiff * radiusDiff;

		if (radiusDiffSqr >= centerDiffSqrLength) {
			if (radiusDiff >= 0.0) {
				load(bv);
				return;
			}
		}

		float centerDiffLength = (float) Math.sqrt(centerDiffSqrLength);
		if (centerDiffLength > 1e-6f) {
			float factor = (centerDiffLength + radiusDiff) / (2.0f * centerDiffLength);
			mCenter.add(centerDiff.mult(factor));
		}

		mRadius = 0.5f * (centerDiffLength + mRadius + bv.mRadius);
	}

	public void setCenter(Float3 center) {
		mCenter = center;
	}

	public void setRadius(float radius) {
		mRadius = radius;
	}

	public BoundingVolume transform(Transformation t) {
		return BoundingVolumes.transform(this, t);
	}

}
