/*******************************************************************************
 * Copyright 2013 Bo Meier (bo.meier@gmail.com)
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 *   http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 ******************************************************************************/
package org.droid3d.scenegraph;

import org.droid3d.math.Float3;
import org.droid3d.math.MathF;
import org.droid3d.math.Matrix44;

public class Camera {
	// viewing frustum
	private float	mNear;
	private float	mFar;
	private float	mBottom;
	private float	mTop;
	private float	mLeft;
	private float	mRight;

	// viewport
	private int		mX;
	private int		mY;
	private int		mWidth;
	private int		mHeight;

	// camera frame
	private Float3	mLocation	= new Float3();
	private Float3	mDirection	= new Float3();
	private Float3	mRVector	= new Float3();
	private Float3	mUVector	= new Float3();

	private boolean	mPerspective;

	/**
	 * Create a perspective camera.
	 */
	public Camera() {
		this(true);
	}

	/**
	 * Create a camera.
	 * 
	 * @param perspective
	 */
	public Camera(boolean perspective) {
		mPerspective = perspective;
		setFrustum(-1f, 1f, -1f, 1f, 1f, 100f);
		lookAt(0f, 0f, 0f, 0f, 0f, -1f, 0f, 1f, 0f);
	}

	/**
	 * Get the projection matrix.
	 * 
	 * @return
	 */
	public Matrix44 getProjectionMatrix() {
		Matrix44 projMatrix = new Matrix44();
		return getProjectionMatrix(projMatrix);
	}

	public Matrix44 getProjectionMatrix(Matrix44 projMatrix) {
		float n = mNear;
		float f = mFar;
		float r = mRight;
		float l = mLeft;
		float t = mTop;
		float b = mBottom;
		if (mPerspective) {
			projMatrix.loadFrustum(l, r, b, t, n, f);
		} else {
			projMatrix.loadOrtho(l, r, b, t, n, f);
		}

		return projMatrix;
	}

	/**
	 * Get the view matrix.
	 * 
	 * @return
	 */
	public Matrix44 getViewMatrix() {
		Matrix44 v = new Matrix44();
		getViewMatrix(v);
		return v;
	}

	public Matrix44 getViewMatrix(Matrix44 view) {
		view.loadView(mLocation, mDirection, mRVector, mUVector);
		return view;
	}

	/**
	 * Get the viewport height.
	 * 
	 * @return
	 */
	public int getViewportHeight() {
		return mHeight;
	}

	/**
	 * Get the viewport width.
	 * 
	 * @return
	 */
	public int getViewportWidth() {
		return mWidth;
	}

	public int getViewportX() {
		return mX;
	}

	public int getViewPortY() {
		return mY;
	}

	/**
	 * @return
	 */
	public boolean isPerspective() {
		return mPerspective;
	}

	/**
	 * @param eyeX
	 * @param eyeY
	 * @param eyeZ
	 * @param centerX
	 * @param centerY
	 * @param centerZ
	 * @param upX
	 * @param upY
	 * @param upZ
	 */
	public void lookAt(float eyeX, float eyeY, float eyeZ, float centerX, float centerY, float centerZ, float upX,
			float upY, float upZ) {
		lookAt(new Float3(eyeX, eyeY, eyeZ), new Float3(centerX, centerY, centerZ), new Float3(upX, upY, upZ));
	}

	/**
	 * @param eye
	 * @param center
	 * @param up
	 */
	public void lookAt(Float3 eye, Float3 center, Float3 up) {
		Float3 d = Float3.sub(center, eye);
		d.normalize();

		Float3 r = Float3.cross(d, up);
		r.normalize();

		Float3 u = Float3.cross(r, d);
		u.normalize();

		setFrame(eye, d, u, r);
	}

	private void setAxes(Float3 dVector, Float3 uVector, Float3 rVector) {
		mDirection.load(dVector);
		mUVector.load(uVector);
		mRVector.load(rVector);
	}

	public void setFrame(Float3 location, Float3 direction, Float3 up, Float3 right) {
		mLocation.load(location);
		setAxes(direction, up, right);
	}

	/**
	 * Set the frustum.
	 * 
	 * @param fov
	 * @param ratio
	 * @param near
	 * @param far
	 */
	public void setFrustum(float fov, float ratio, float near, float far) {
		float halfFOVRad = 0.5f * fov * MathF.DEG_2_RAD;
		mTop = (float) (near * Math.tan(halfFOVRad));
		mRight = ratio * mTop;
		mBottom = -mTop;
		mLeft = -mRight;
		mNear = near;
		mFar = far;
	}

	/**
	 * Set the frustum.
	 * 
	 * @param left
	 * @param right
	 * @param bottom
	 * @param top
	 * @param near
	 * @param far
	 */
	public void setFrustum(float left, float right, float bottom, float top, float near, float far) {
		mLeft = left;
		mRight = right;
		mBottom = bottom;
		mTop = top;
		mNear = near;
		mFar = far;
	}

	/**
	 * Set the camera location.
	 * 
	 * @param x
	 * @param y
	 * @param z
	 */
	public void setLocation(float x, float y, float z) {
		mLocation.load(x, y, z);
	}

	/**
	 * Set the camera location.
	 * 
	 * @param location
	 */
	public void setLocation(Float3 location) {
		mLocation.load(location);
	}

	/**
	 * Set the viewport.
	 * 
	 * @param x
	 * @param y
	 * @param width
	 * @param height
	 */
	public void setViewport(int x, int y, int width, int height) {
		mX = x;
		mY = y;
		mWidth = width;
		mHeight = height;
	}

}
