/*******************************************************************************
 * Copyright 2013 Bo Meier (bo.meier@gmail.com)
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 *   http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 ******************************************************************************/
package org.droid3d.scenegraph;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;

import org.droid3d.renderer.Light;
import org.droid3d.renderer.states.RenderState;
import org.droid3d.renderer.states.RenderState.Type;

public abstract class Node extends SceneObject {
	public enum CullingMode {
		DYNAMIC, ALWAYS, NEVER;
	}

	private GroupNode									mParent;
	private final Map<RenderState.Type, RenderState>	mStates					= new HashMap<RenderState.Type, RenderState>();
	private List<Light>									mLights					= new ArrayList<Light>();

	private CullingMode									mCullingMode			= CullingMode.DYNAMIC;

	private Transformation								mLocalTransformation	= new Transformation();
	private Transformation								mWorldTransformation	= new Transformation();
	public boolean										worldIsCurrent			= false;

	private BoundingVolume								mBoundingVolume			= new BoundingVolume();
	public boolean										worldBoundIsCurrent		= false;

	protected Node() {
	}

	public void addLight(Light light) {
		mLights.add(light);
	}

	public void addState(RenderState state) {
		mStates.put(state.getType(), state);
	}

	public abstract void cull(Culler culler, boolean doCull);

	/**
	 * Get the bounding volume.
	 * 
	 * @return
	 */
	public BoundingVolume getBoundingVolume() {
		return mBoundingVolume;
	}

	public CullingMode getCullingMode() {
		return mCullingMode;
	}

	/**
	 * Get the number of lights attached to the node.
	 * 
	 * @return
	 */
	public int getLightCount() {
		return mLights.size();
	}

	public List<Light> getLights() {
		return mLights;
	}

	public Transformation getLocalTransformation() {
		return mLocalTransformation;
	}

	public final GroupNode getParent() {
		return mParent;
	}

	public RenderState getState(RenderState.Type type) {
		return mStates.get(type);
	}

	public Transformation getWorldTransformation() {
		return mWorldTransformation;
	}

	public void onCull(Culler culler, boolean doCull) {
		if (getCullingMode() == CullingMode.ALWAYS) {
			return;
		}

		if (getCullingMode() == CullingMode.NEVER) {
			doCull = false;
		}

		if (!doCull || culler.isVisible(getBoundingVolume())) {
			onGetVisibleUpdateControllers(culler.getCamera());
			cull(culler, doCull);
		}
	}

	private void popState(Map<RenderState.Type, LinkedList<RenderState>> stateStacks, LinkedList<Light> lightStack) {
		for (RenderState.Type type : mStates.keySet()) {
			LinkedList<RenderState> stack = stateStacks.get(type);
			stack.removeLast();
		}

		for (int i = 0; i < mLights.size(); i++) {
			lightStack.removeLast();
		}

	}

	protected final void propagateBoundToRoot() {
		if (mParent != null) {
			mParent.updateBoundingVolume();
			mParent.propagateBoundToRoot();
		}
	}

	protected final void propagateStateFromRoot(Map<Type, LinkedList<RenderState>> stateStacks,
			LinkedList<Light> lightStack) {
		if (mParent != null) {
			mParent.propagateStateFromRoot(stateStacks, lightStack);
		}

		pushState(stateStacks, lightStack);
	}

	private void pushState(Map<RenderState.Type, LinkedList<RenderState>> stateStacks, LinkedList<Light> lightStack) {
		for (RenderState.Type type : mStates.keySet()) {
			LinkedList<RenderState> stateStack = stateStacks.get(type);
			stateStack.add(mStates.get(type));
		}

		for (Light light : mLights) {
			lightStack.add(light);
		}
	}

	public void removeState(RenderState.Type type) {
		for (int i = 0; i < mStates.size(); i++) {
			if (mStates.get(i).getType() == type) {
				mStates.remove(i);
				return;
			}
		}
	}

	/**
	 * Set the bounding volume.
	 * 
	 * @param boundingVolume
	 */
	public void setBoundingVolume(BoundingVolume boundingVolume) {
		mBoundingVolume = boundingVolume;
	}

	public void setCullingMode(CullingMode culling) {
		mCullingMode = culling;
	}

	public void setLocalTransformation(Transformation localTransformation) {
		mLocalTransformation = localTransformation;
	}

	public final void setParent(GroupNode parent) {
		mParent = parent;
	}

	public void setWorldTransformation(Transformation worldTransformation) {
		mWorldTransformation = worldTransformation;
	}

	/**
	 * Update the bounding volume.
	 */
	public abstract void updateBoundingVolume();

	public final void updateBS() {
		updateBoundingVolume();
		propagateBoundToRoot();
	}

	public final void updateGS() {
		updateGS(Float.MIN_VALUE, true);
	}

	public final void updateGS(float time) {
		updateGS(time, true);
	}

	public final void updateGS(float time, boolean isInitiator) {
		updateWorldData(time);
		updateBoundingVolume();
		if (isInitiator) {
			propagateBoundToRoot();
		}
	}

	public void updateRS() {
		updateRS(null, null);
	}

	protected void updateRS(Map<RenderState.Type, LinkedList<RenderState>> stateStacks, LinkedList<Light> lightStack) {
		boolean isInitiator = stateStacks == null;
		int numStates = mStates.size();

		if (isInitiator) {
			stateStacks = new HashMap<RenderState.Type, LinkedList<RenderState>>();
			for (RenderState state : RenderState.DEFAULT) {
				LinkedList<RenderState> stack = new LinkedList<RenderState>();
				stack.add(state);
				stateStacks.put(state.getType(), stack);
			}
			lightStack = new LinkedList<Light>();
			propagateStateFromRoot(stateStacks, lightStack);
		} else {
			pushState(stateStacks, lightStack);
		}

		updateState(stateStacks, lightStack);
		if (!isInitiator && numStates > 0) {
			popState(stateStacks, lightStack);
		}
	}

	protected abstract void updateState(Map<Type, LinkedList<RenderState>> stateStacks, LinkedList<Light> lights);

	protected void updateWorldData(float time) {
		updateControllers(time);
		if (!worldIsCurrent) {
			if (mParent != null) {
				mWorldTransformation.product(mParent.getWorldTransformation(), mLocalTransformation);
			} else {
				mWorldTransformation = mLocalTransformation;
			}
		}
	}
}
