/*******************************************************************************
 * Copyright 2013 Bo Meier (bo.meier@gmail.com)
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 *   http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 ******************************************************************************/
package org.droid3d.scenegraph;

import org.droid3d.math.Float2;
import org.droid3d.math.Float3;
import org.droid3d.renderer.Color;
import org.droid3d.renderer.Font;
import org.droid3d.renderer.Material;
import org.droid3d.renderer.Material.BlendMode;
import org.droid3d.renderer.Mesh;
import org.droid3d.renderer.RenderMode;
import org.droid3d.renderer.Renderer;
import org.droid3d.renderer.buffer.Buffer;
import org.droid3d.renderer.buffer.IndexBuffer;
import org.droid3d.renderer.buffer.VertexBuffer;
import org.droid3d.utils.buffer.Channel;
import org.droid3d.utils.buffer.ChannelFactory;
import org.droid3d.utils.buffer.VertexAttributeFactory;

public class Text extends Geometry {
	private Font			mFont;
	private int				mMaxLength;
	private float			mCursorX		= 0f;
	private float			mCursorY		= 0f;
	private float			mLineSpacing	= 1.0f;
	private int				mLength			= 0;

	private VertexBuffer	mVBuffer;
	private IndexBuffer		mIBuffer;
	private Mesh			mMesh;

	private Channel<Float3>	mPositionChannel;
	private Channel<Float2>	mTexCoordChannel;
	private Channel<Color>	mColorChannel;

	public Text(Font font, int maxLength) {
		mMaxLength = maxLength;
		mFont = font;

		mVBuffer = new VertexBuffer(4 * maxLength, Buffer.Usage.DYNAMIC,
				VertexAttributeFactory.newPositionAttr(), VertexAttributeFactory.newColorAttr(),
				VertexAttributeFactory.newTexCoordAttr());

		mPositionChannel = ChannelFactory.newPositionChannel(mVBuffer);
		mTexCoordChannel = ChannelFactory.newTexCoordChannel(mVBuffer);
		mColorChannel = ChannelFactory.newColorChannel(mVBuffer);

		mIBuffer = new IndexBuffer(6 * maxLength, Buffer.Usage.DYNAMIC);
		mMesh = new Mesh(mVBuffer, mIBuffer, RenderMode.TRIANGLES);
		setMesh(mMesh);

		Material mat = new Material();
		mat.addTexture(mFont.getTexture(), BlendMode.MODULATE);
		setMaterial(mat);

	}

	public void append(String text) {
		append(text, Color.WHITE);
	}

	public void append(String text, Color color) {

		for (int i = 0; i < text.length(); i++) {
			char c = text.charAt(i);
			if (c == '\n') {
				mCursorY -= (mFont.getLineHeight() * mLineSpacing);
				mCursorX = 0f;
				continue;
			}

			if (mLength + 1 >= mMaxLength) {
				break;
			}

			Font.Glyph glyph = mFont.getGlyph(c);

			float x = mCursorX + glyph.offset[0];
			float y = mCursorY - glyph.offset[1]; // -
													// glyph.offset[1];
			int offset = mLength * 4;
			mPositionChannel.write(offset, new Float3(x + glyph.width, y, 0));
			mPositionChannel.write(offset + 1, new Float3(x, y, 0));
			mPositionChannel.write(offset + 2, new Float3(x, y - glyph.height, 0));
			mPositionChannel.write(offset + 3, new Float3(x + glyph.width, y - glyph.height, 0));
			mColorChannel.write(offset + 0, color);
			mColorChannel.write(offset + 1, color);
			mColorChannel.write(offset + 2, color);
			mColorChannel.write(offset + 3, color);
			mTexCoordChannel.write(offset + 0, glyph.uvs[0]);
			mTexCoordChannel.write(offset + 1, glyph.uvs[1]);
			mTexCoordChannel.write(offset + 2, glyph.uvs[2]);
			mTexCoordChannel.write(offset + 3, glyph.uvs[3]);

			mIBuffer.setIndex((mLength * 6), new short[] {// @formatter:off
					(short) offset, (short) (1 + offset), (short) (2 + offset),
							(short) (0 + offset), (short) (2 + offset),
							(short) (3 + offset) });// @formatter:on

			mCursorX += glyph.xadvance;
			mLength++;
		}
		mMesh.setIndexCount(mLength * 6);

	}

	public void clear() {
		mMesh.setIndexCount(0);
		mCursorX = 0f;
		mCursorY = 0f;
		mLength = 0;
	}

	public void update(Renderer renderer) {
		renderer.update(mMesh.getIndexBuffer());
		renderer.update(mMesh.getVertexBuffer());
	}
}
