/*******************************************************************************
 * Copyright 2013 Bo Meier (bo.meier@gmail.com)
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 *   http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 ******************************************************************************/
package org.droid3d.scenegraph;

import org.droid3d.math.Matrix44;
import org.droid3d.math.Float3;

/**
 * 
 */
public class Transformation {

	private Matrix44	mMatrix;
	private Float3	mScale;

	private boolean		mIdentity;
	private boolean		mRSMatrix;
	private boolean		mUniformScale;

	public Transformation() {
		mIdentity = true;
		mRSMatrix = true;
		mUniformScale = true;

		mMatrix = new Matrix44();
		mMatrix.loadIdentity();
		mScale = new Float3(1f, 1f, 1f);
	}

	protected Transformation(Transformation pTransformation) {
		load(pTransformation);
	}

	public Float3 applyForward(Float3 v) {
		if (isIdentity()) {
			return new Float3(v);
		}

		Float3 translate = getTranslate();
		if (isRSMatrix()) {
			Float3 output = new Float3(v);
			output.x *= mScale.x;
			output.y *= mScale.y;
			output.z *= mScale.z;

			output = mMatrix.mult(output);
			output.add(translate);
			return output;
		}

		Float3 output = new Float3(v);
		output = mMatrix.mult(output);
		output.add(translate);
		return output;
	}

	public Matrix44 getMatrix() {
		return mMatrix;
	}

	public float getNorm() {
		if (isRSMatrix()) {
			float max = Math.abs(mScale.x);

			float sYAbs = Math.abs(mScale.y);
			if (sYAbs > max) {
				max = sYAbs;
			}

			float sZAbs = Math.abs(mScale.z);
			if (sZAbs > max) {
				max = sZAbs;
			}

			return max;
		}

		float maxRowSum = Math.abs(mMatrix.get(0, 0))
				+ Math.abs(mMatrix.get(0, 1)) + Math.abs(mMatrix.get(0, 2));
		float rowSum = Math.abs(mMatrix.get(1, 0))
				+ Math.abs(mMatrix.get(1, 1)) + Math.abs(mMatrix.get(1, 2));
		if (rowSum > maxRowSum) {
			maxRowSum = rowSum;
		}

		rowSum = Math.abs(mMatrix.get(2, 0)) + Math.abs(mMatrix.get(2, 1))
				+ Math.abs(mMatrix.get(2, 2));
		if (rowSum > maxRowSum) {
			maxRowSum = rowSum;
		}
		return maxRowSum;
	}

	public Float3 getScale() {
		return mScale;
	}

	public Matrix44 getTransformation(Matrix44 matrix) {
		if (matrix == null) {
			matrix = new Matrix44();
		}

		matrix.load(mMatrix);
		if (mRSMatrix) {
			matrix.scale(mScale.x, mScale.y, mScale.z);
		}
		return matrix;
	}

	public Float3 getTranslate() {
		Float3 t = new Float3();
		t.x = mMatrix.get(0, 3);
		t.y = mMatrix.get(1, 3);
		t.z = mMatrix.get(2, 3);
		return t;
	}

	public float getUniformScale() {
		assert mRSMatrix && mUniformScale;
		return mScale.x;
	}

	public boolean isIdentity() {
		return mIdentity;
	}

	public boolean isRSMatrix() {
		return mRSMatrix;
	}

	public boolean isUniformScale() {
		return mRSMatrix && mUniformScale;
	}

	private void load(Transformation t) {
		mIdentity = t.mIdentity;
		mRSMatrix = t.mRSMatrix;
		mUniformScale = t.mUniformScale;

		mMatrix = new Matrix44(t.mMatrix);
		mScale = new Float3(t.mScale);
	}

	public void loadIdentity() {
		mMatrix.loadIdentity();
		mScale.load(1f, 1f, 1f);
		mIdentity = true;
		mRSMatrix = true;
		mUniformScale = true;
	}

	public void product(Transformation pA, Transformation pB) {
		if (pA.isIdentity()) {
			load(pB);
			return;
		}

		if (pB.isIdentity()) {
			load(pA);
			return;
		}

		if (pA.isRSMatrix() && pB.isRSMatrix()) {
			if (pA.isUniformScale()) {
				Matrix44 rotate = Matrix44.mult(pA.mMatrix, pB.mMatrix);
				setRotate(rotate);

				Float3 transA = pA.getTranslate();
				Float3 transB = pB.getTranslate();

				Float3 translate = pA.mMatrix.mult(transB)
						.mult(pA.getUniformScale()).add(transA);

				setTranslate(translate);

				if (pB.isUniformScale()) {
					setScale(pA.getUniformScale() * pB.getUniformScale());
				} else {
					Float3 scale = new Float3(pB.getScale());
					scale.mult(pA.getUniformScale());
					setScale(scale);
				}
				return;
			}
		}

		Matrix44 a = pA.getTransformation(null);
		Matrix44 b = pB.getTransformation(null);
		setTransformation(Matrix44.mult(a, b));
		Float3 translate = a.mult(pB.getTranslate()).add(pA.getTranslate());
		setTranslate(translate);
	}

	public void setRotate(Matrix44 r) {
		Float3 translate = getTranslate();
		mMatrix.load(r);
		setTranslate(translate);
		mIdentity = false;
		mRSMatrix = true;
	}

	public void setScale(float x, float y, float z) {
		mScale.load(x, y, z);
		mIdentity = false;
		mUniformScale = false;
	}

	public void setScale(Float3 scale) {
		mScale.load(scale);
		mIdentity = false;
		mUniformScale = false;
	}

	public void setTransformation(Matrix44 t) {
		mMatrix.load(t);
		mIdentity = false;
		mRSMatrix = false;
		mUniformScale = false;
	}

	public void setTranslate(float x, float y, float z) {
		mIdentity = false;
		mMatrix.set(0, 3, x);
		mMatrix.set(1, 3, y);
		mMatrix.set(2, 3, z);
	}

	public void setTranslate(Float3 t) {
		mIdentity = false;
		mMatrix.set(0, 3, t.x);
		mMatrix.set(1, 3, t.y);
		mMatrix.set(2, 3, t.z);
	}

	public void setScale(float scale) {
		mScale.load(scale, scale, scale);
		mIdentity = false;
		mUniformScale = true;
	}
}
