/*******************************************************************************
 * Copyright 2013 Bo Meier (bo.meier@gmail.com)
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 *   http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 ******************************************************************************/
package org.droid3d.utils;

import java.util.ArrayList;
import java.util.List;

import org.droid3d.math.Float2;
import org.droid3d.math.Float3;
import org.droid3d.math.MathF;
import org.droid3d.renderer.Color;
import org.droid3d.renderer.Mesh;
import org.droid3d.renderer.RenderMode;
import org.droid3d.renderer.buffer.IndexBuffer;
import org.droid3d.renderer.buffer.VertexAttribute;
import org.droid3d.renderer.buffer.VertexBuffer;
import org.droid3d.utils.buffer.Channel;
import org.droid3d.utils.buffer.ChannelFactory;
import org.droid3d.utils.buffer.VertexAttributeFactory;

/**
 * Create meshes.
 */
public class MeshFactory {

	public static Mesh createAxes(float extent) {
		VertexBuffer vBuffer = new VertexBuffer(6, VertexAttributeFactory.newPositionAttr(),
				VertexAttributeFactory.newColorAttr());
		Channel<Float3> pos = ChannelFactory.newPositionChannel(vBuffer);
		pos.write(0, new Float3(0f, 0f, 0f));
		pos.write(1, new Float3(extent, 0f, 0f));
		pos.write(2, new Float3(0f, 0f, 0f));
		pos.write(3, new Float3(0f, extent, 0f));
		pos.write(4, new Float3(0f, 0f, 0f));
		pos.write(5, new Float3(0f, 0f, extent));

		Channel<Color> col = ChannelFactory.newColorChannel(vBuffer);
		col.write(0, Color.RED);
		col.write(1, Color.RED);
		col.write(2, Color.GREEEN);
		col.write(3, Color.GREEEN);
		col.write(4, Color.BLUE);
		col.write(5, Color.BLUE);

		IndexBuffer iBuffer = new IndexBuffer(new short[] { 0, 1, 2, 3, 4, 5 });
		return new Mesh(vBuffer, iBuffer, RenderMode.LINES);
	}

	public static Mesh createCircle(float r, int s) {
		float dphi = 2 * MathF.PI / s;

		VertexAttribute[] attr = new VertexAttribute[] { VertexAttributeFactory.newPositionAttr(),
				VertexAttributeFactory.newNormalAttr() };
		VertexBuffer vBuffer = new VertexBuffer(s + 2, attr);

		Channel<Float3> posCh = ChannelFactory.newPositionChannel(vBuffer);
		Channel<Float3> normCh = ChannelFactory.newNormalChannel(vBuffer);
		Float3 n = new Float3(0f, 1f, 0f);
		posCh.write(0, new Float3(0f, 0f, 0f));
		normCh.write(0, n);
		for (int i = 0; i < s; i++) {
			float phi = dphi * i;
			float cp = MathF.cos(phi);
			float sp = MathF.sin(phi);
			posCh.write(i + 1, new Float3(r * cp, 0f, r * sp));
			normCh.write(i + 1, n);
		}
		posCh.write(s + 1, new Float3(r, 0f, 0f));
		normCh.write(s + 1, n);

		short[] indices = new short[s * 3];
		int idx = 0;
		for (int i = 0; i < s; i++, idx += 3) {
			indices[idx] = 0;
			indices[idx + 1] = (short) (i + 2);
			indices[idx + 2] = (short) (i + 1);
		}
		IndexBuffer iBuffer = new IndexBuffer(indices);

		Mesh mesh = new Mesh(vBuffer, iBuffer);
		return mesh;
	}

	public static Mesh createCone(float h, float r, int s) {
		float dphi = 2 * MathF.PI / s;

		VertexAttribute[] attr = new VertexAttribute[] { VertexAttributeFactory.newPositionAttr(),
				VertexAttributeFactory.newNormalAttr() };
		int sideVertices = 2 * (s + 1);
		int botVertices = s + 2;

		VertexBuffer vBuffer = new VertexBuffer(sideVertices + botVertices, attr);
		Channel<Float3> posCh = ChannelFactory.newPositionChannel(vBuffer);
		Channel<Float3> normCh = ChannelFactory.newNormalChannel(vBuffer);

		// sides
		int idx = 0;
		for (int i = 0; i <= s; i++, idx += 2) {
			float phi = dphi * i;
			float cp = MathF.cos(phi);
			float sp = MathF.sin(phi);

			float x = r * sp;
			float z = r * cp;
			// tangent at base
			Float3 t = new Float3(cp, -sp, 0f);
			Float3 up = new Float3(-x, h, -z).normalize();
			Float3 n = Float3.cross(t, up).normalize();

			posCh.write(idx, new Float3(0f, h, 0f));
			normCh.write(idx, n);
			posCh.write(idx + 1, new Float3(r * sp, 0f, r * cp));
			normCh.write(idx + 1, n);
		}

		// bot
		idx = sideVertices;
		posCh.write(idx, new Float3(0f, 0f, 0f));
		normCh.write(idx, new Float3(0f, -1f, 0f));
		idx += 1;
		for (int i = 0; i <= s; i++, idx++) {
			float phi = dphi * i;
			float cp = MathF.cos(phi);
			float sp = MathF.sin(phi);
			posCh.write(idx, new Float3(r * sp, 0f, r * cp));
			normCh.write(idx, new Float3(0, -1f, 0));
		}

		short[] indices = new short[2 * s * 3];

		// side
		idx = 0;
		for (int i = 0; i < s; i++) {
			int offset = i * 2;
			indices[idx++] = (short) offset;
			indices[idx++] = (short) (offset + 1);
			indices[idx++] = (short) (offset + 3);
		}

		// bot
		for (int i = 0; i < s; i++) {
			int offset = sideVertices + 1 + i;
			indices[idx++] = (short) sideVertices;
			indices[idx++] = (short) (offset + 1);
			indices[idx++] = (short) (offset);
		}

		IndexBuffer iBuffer = new IndexBuffer(indices);

		Mesh mesh = new Mesh(vBuffer, iBuffer);
		return mesh;
	}

	/**
	 * Create a cube.
	 * 
	 * @param extent
	 * @param createNormals
	 * @param createTexCoords
	 * @return
	 */
	public static Mesh createCube(float extent, boolean createNormals, boolean createTexCoords) {

		final Float3[] positions = new Float3[] {// @formatter:off
				// front
				new Float3(extent, extent, extent),
				new Float3(-extent, extent, extent),
				new Float3(-extent, -extent, extent),
				new Float3(extent, -extent, extent),

				// top
				new Float3(extent, extent, -extent),
				new Float3(-extent, extent, -extent),
				new Float3(-extent, extent, extent),
				new Float3(extent, extent, extent),

				// right
				new Float3(extent, extent, -extent),
				new Float3(extent, extent, extent),
				new Float3(extent, -extent, extent),
				new Float3(extent, -extent, -extent),

				// back
				new Float3(extent, extent, -extent),
				new Float3(extent, -extent, -extent),
				new Float3(-extent, -extent, -extent),
				new Float3(-extent, extent, -extent),

				// bottom
				new Float3(extent, -extent, -extent),
				new Float3(-extent, -extent, extent),
				new Float3(-extent, -extent, -extent),
				new Float3(extent, -extent, extent),

				// left
				new Float3(-extent, extent, -extent),
				new Float3(-extent, -extent, extent),
				new Float3(-extent, extent, extent),
				new Float3(-extent, -extent, -extent) };// @formatter:on

		final Color[] colors = new Color[] {// @formatter:off
		new Color(1f, 0f, 0f, 1f), new Color(0f, 1f, 0f, 1f),
				new Color(0f, 0f, 1f, 1f), new Color(1f, 1f, 0f, 1f),
				new Color(1f, 0f, 1f, 1f), new Color(0f, 1f, 1f, 1f) };// @formatter:on

		List<VertexAttribute> attributes = new ArrayList<VertexAttribute>();
		attributes.add(VertexAttributeFactory.newPositionAttr());
		attributes.add(VertexAttributeFactory.newColorAttr());
		if (createNormals) {
			attributes.add(VertexAttributeFactory.newNormalAttr());
		}

		if (createTexCoords) {
			attributes.add(VertexAttributeFactory.newTexCoordAttr());
		}
		int vertexCount = positions.length;
		VertexBuffer vBuffer = new VertexBuffer(vertexCount, attributes);

		Channel<Float3> posChannel = ChannelFactory.newPositionChannel(vBuffer);
		Channel<Color> colChannel = ChannelFactory.newColorChannel(vBuffer);
		for (int i = 0; i < vertexCount; i++) {
			posChannel.write(i, positions[i]);
			colChannel.write(i, colors[i / 4]);
		}

		if (createNormals) {
			final Float3[] normals = new Float3[] {// @formatter:off
			// front
					new Float3(0f, 0f, 1f),
					// top
					new Float3(0f, 1f, 0f),
					// right
					new Float3(1f, 0f, 0f),
					// back
					new Float3(0f, 0f, -1f),
					// bottom
					new Float3(0f, -1f, 0f),
					// left
					new Float3(-1f, 0f, 0f), };// @formatter:on

			Channel<Float3> normalChannel = ChannelFactory.newNormalChannel(vBuffer);
			for (int i = 0; i < vertexCount; i++) {
				normalChannel.write(i, normals[i / 6]);
			}
		}

		if (createTexCoords) {
			Float2[] texCoords = new Float2[] {// @formatter:off
			new Float2(1f, 1f), new Float2(0f, 1f), new Float2(0f, 0f),
					new Float2(1f, 0f),

					new Float2(1f, 1f), new Float2(0f, 1f),
					new Float2(0f, 0f), new Float2(1f, 0f),

					new Float2(1f, 1f), new Float2(0f, 1f),
					new Float2(0f, 0f), new Float2(1f, 0f),

					new Float2(0f, 1f), new Float2(0f, 0f),
					new Float2(1f, 0f), new Float2(1f, 1f),

					new Float2(0f, 1f), new Float2(1f, 0f),
					new Float2(1f, 1f), new Float2(0f, 0f),

					new Float2(0f, 1f), new Float2(1f, 0f),
					new Float2(1f, 1f), new Float2(0f, 0f) }; // @formatter:on

			Channel<Float2> texCoordChannel = ChannelFactory.newTexCoordChannel(vBuffer);

			for (int i = 0; i < vertexCount; i++) {
				texCoordChannel.write(i, texCoords[i]);
			}
		}

		final short[] indices = new short[] {// @formatter:off
		0, 1, 2, 0, 2, 3, // front
				4, 5, 6, 4, 6, 7, // top
				8, 9, 10, 8, 10, 11, // right
				12, 13, 14, 12, 14, 15, // back
				16, 17, 18, 16, 19, 17, // bottom
				20, 21, 22, 20, 23, 21 // left
		};// @formatter:on

		IndexBuffer iBuffer = new IndexBuffer(indices);

		return new Mesh(vBuffer, iBuffer);
	}

	/**
	 * Create a cube.
	 * 
	 * @param extent
	 * @return
	 */
	public static Mesh createCube8(float extent) {

		final Float3[] positions = new Float3[] {
				// @formatter:off
				new Float3(extent, extent, extent),
				new Float3(-extent, extent, extent),
				new Float3(-extent, -extent, extent),
				new Float3(extent, -extent, extent),

				new Float3(extent, extent, -extent),
				new Float3(-extent, extent, -extent),
				new Float3(-extent, -extent, -extent),
				new Float3(extent, -extent, -extent)
		// @formatter:on
		};

		final Color[] colors = new Color[] {
				// @formatter:off
				new Color(1f, 0f, 0f, 0.5f), new Color(0f, 1f, 0f, 0.5f),
				new Color(0f, 0f, 1f, 0.5f), new Color(1f, 1f, 0f, 0.5f),
				new Color(1f, 0f, 1f, 0.5f), new Color(0f, 1f, 1f, 0.5f),
				new Color(0f, 0f, 0f, 0.5f), new Color(1f, 1f, 1f, 0.5f)
		// @formatter:on
		};

		final int numVertices = positions.length;
		VertexBuffer vBuffer = new VertexBuffer(numVertices, VertexAttributeFactory.newPositionAttr(),
				VertexAttributeFactory.newColorAttr());

		Channel<Float3> posChannel = ChannelFactory.newPositionChannel(vBuffer);
		for (int i = 0; i < numVertices; i++) {
			posChannel.write(i, positions[i]);
		}

		Channel<Color> colChannel = ChannelFactory.newColorChannel(vBuffer);
		for (int i = 0; i < numVertices; i++) {
			Color color = colors[i];
			colChannel.write(i, color);
		}

		final short[] indices = new short[] {// @formatter:off
		0, 1, 2, 0, 2, 3, // front
				4, 5, 1, 4, 1, 0, // top
				4, 0, 3, 4, 3, 7, // right
				4, 6, 5, 4, 7, 6, // back
				7, 2, 6, 7, 3, 2, // bottom
				5, 2, 1, 5, 6, 2 // left
		};// @formatter:on
		IndexBuffer iBuffer = new IndexBuffer(indices);
		return new Mesh(vBuffer, iBuffer);
	}

	public static Mesh createCylinder(float h, float r, int s) {

		VertexAttribute[] attr = new VertexAttribute[] {
//@formatter:off
				VertexAttributeFactory.newPositionAttr(),
				VertexAttributeFactory.newNormalAttr() 
//@formatter:on
		};
		int sideVertices = 2 * (s + 1);
		int capVertices = s + 2;
		int numVertices = sideVertices + 2 * capVertices;
		VertexBuffer vBuffer = new VertexBuffer(numVertices, attr);
		Channel<Float3> pos = ChannelFactory.newPositionChannel(vBuffer);
		Channel<Float3> norm = ChannelFactory.newNormalChannel(vBuffer);

		float dphi = 2 * MathF.PI / s;
		Float3 n = new Float3(0f, 1f, 0f);
		float top = h / 2;
		float bot = -top;

		int idx = 0;
		int topOffset = sideVertices;
		int botOffset = sideVertices + capVertices;

		// center of top/bot cap
		pos.write(botOffset - 1, new Float3(0f, top, 0f));
		norm.write(botOffset - 1, new Float3(0f, 1f, 0f));
		pos.write(numVertices - 1, new Float3(0f, bot, 0f));
		norm.write(numVertices - 1, new Float3(0f, -1f, 0f));

		for (int i = 0; i < s; i++, idx += 2) {
			float phi = dphi * i;
			float cp = MathF.cos(phi);
			float sp = -MathF.sin(phi);

			float x = r * cp;
			float z = r * sp;

			// side
			n.set(cp, 0f, sp);
			pos.write(idx, new Float3(x, top, z));
			norm.write(idx, n);
			pos.write(idx + 1, new Float3(x, bot, z));
			norm.write(idx + 1, n);

			// top
			pos.write(i + topOffset, new Float3(x, top, z));
			norm.write(i + topOffset, new Float3(0f, 1f, 0f));

			// bot
			pos.write(i + botOffset, new Float3(x, bot, z));
			norm.write(i + botOffset, new Float3(0f, -1f, 0f));
		}

		// duplicate 1st/2nd vertex
		n.load(1f, 0f, 0f);
		pos.write(2 * s, new Float3(r, top, 0f));
		norm.write(2 * s, n);
		pos.write(2 * s + 1, new Float3(r, bot, 0f));
		norm.write(2 * s + 1, n);

		pos.write(botOffset - 2, new Float3(r, top, 0f));
		norm.write(botOffset - 2, new Float3(0f, 1f, 0f));

		// bot
		pos.write(numVertices - 2, new Float3(r, bot, 0f));
		norm.write(numVertices - 2, new Float3(0f, -1f, 0f));

		idx = 0;
		int sideIndices = 6 * s;
		int capIndices = 3 * s;
		short[] indices = new short[sideIndices + 2 * capIndices];
		for (int i = 0; i < s; i++) {
			// side
			int offset = i * 2;
			indices[idx++] = (short) offset;
			indices[idx++] = (short) (offset + 1);
			indices[idx++] = (short) (offset + 2);

			indices[idx++] = (short) (offset + 2);
			indices[idx++] = (short) (offset + 1);
			indices[idx++] = (short) (offset + 3);
		}

		// top
		for (int i = 0; i < s; i++) {
			int offset = topOffset + i;
			indices[idx++] = (short) offset;
			indices[idx++] = (short) (offset + 1);
			indices[idx++] = (short) (botOffset - 1);
		}

		// bot
		for (int i = 0; i < s; i++) {
			int offset = botOffset + i;
			indices[idx++] = (short) offset;
			indices[idx++] = (short) (numVertices - 1);
			indices[idx++] = (short) (offset + 1);
		}

		IndexBuffer iBuffer = new IndexBuffer(indices);
		Mesh mesh = new Mesh(vBuffer, iBuffer, RenderMode.TRIANGLES);
		return mesh;
	}

	public static Mesh createSphere(float radius, int rings, int segments, Color color) {

		List<VertexAttribute> attributes = new ArrayList<VertexAttribute>();
		attributes.add(VertexAttributeFactory.newPositionAttr());
		attributes.add(VertexAttributeFactory.newNormalAttr());
		attributes.add(VertexAttributeFactory.newTexCoordAttr());
		if (color != null) {
			attributes.add(VertexAttributeFactory.newColorAttr());
		}
		int vertexCount = (rings + 1) * (segments + 1);
		VertexBuffer vBuffer = new VertexBuffer(vertexCount, attributes);

		int indexCount = 6 * rings * (segments + 1);
		IndexBuffer iBuffer = new IndexBuffer(indexCount);

		Channel<Float3> posChannel = ChannelFactory.newPositionChannel(vBuffer);
		Channel<Float3> normalChannel = ChannelFactory.newNormalChannel(vBuffer);
		Channel<Float2> texCoordChannel = ChannelFactory.newTexCoordChannel(vBuffer);
		Channel<Color> colorChannel = ChannelFactory.newColorChannel(vBuffer);

		double deltaRingAngle = Math.PI / rings;
		double deltaSegAngle = 2 * Math.PI / segments;
		int vertex = 0;

		for (int ring = 0; ring <= rings; ring++) {
			double r0 = radius * Math.sin(ring * deltaRingAngle);
			double y0 = radius * Math.cos(ring * deltaRingAngle);

			for (int seg = 0; seg <= segments; seg++) {
				double x0 = r0 * Math.sin(seg * deltaSegAngle);
				double z0 = r0 * Math.cos(seg * deltaSegAngle);

				Float3 position = new Float3((float) x0, (float) y0, (float) z0);
				posChannel.write(vertex, position);

				Float3 normal = Float3.normalize(position);
				normalChannel.write(vertex, normal);

				float u = (float) seg / (float) segments;
				float v = (float) ring / (float) rings;
				texCoordChannel.write(vertex, new Float2(u, v));

				if (color != null) {
					colorChannel.write(vertex, color);
				}

				if (ring != rings) {
					iBuffer.add((short) (vertex + segments + 1));
					iBuffer.add((short) (vertex));
					iBuffer.add((short) (vertex + segments));
					iBuffer.add((short) (vertex + segments + 1));
					iBuffer.add((short) (vertex + 1));
					iBuffer.add((short) (vertex));

				}
				vertex++;
			}

		}

		return new Mesh(vBuffer, iBuffer);
	}

	public VertexBuffer merge(VertexBuffer a, VertexBuffer b) {

		return null;
	}
}
