/*******************************************************************************
 * Copyright 2013 Bo Meier (bo.meier@gmail.com)
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 *   http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 ******************************************************************************/
package org.droid3d.math;

import java.util.Random;

import junit.framework.Assert;

import org.junit.Before;
import org.junit.Test;

public class Vector3FTest {
	private Random	mRandom;

	@Before
	public void init()
	{
		mRandom = new Random(System.currentTimeMillis());
	}

	@Test
	public void testAdd()
	{
		float[] tmp1 = new float[] { mRandom.nextFloat(), mRandom.nextFloat(), mRandom.nextFloat() };
		float[] tmp2 = new float[] { mRandom.nextFloat(), mRandom.nextFloat(), mRandom.nextFloat() };

		Float3 v1 = new Float3(tmp1);
		Float3 v2 = new Float3(tmp2);

		Float3 v3 = Float3.add(v1, v2);
		Assert.assertEquals(tmp1[0] + tmp2[0], v3.x);
		Assert.assertEquals(tmp1[1] + tmp2[1], v3.y);
		Assert.assertEquals(tmp1[2] + tmp2[2], v3.z);

		v1.add(v2);
		Assert.assertEquals(tmp1[0] + tmp2[0], v1.x);
		Assert.assertEquals(tmp1[1] + tmp2[1], v1.y);
		Assert.assertEquals(tmp1[2] + tmp2[2], v1.z);

	}

	@Test
	public void testConst()
	{
		float[] tmp = new float[] { mRandom.nextFloat(), mRandom.nextFloat(), mRandom.nextFloat() };

		Float3 v = new Float3(tmp);
		Assert.assertEquals(tmp[0], v.x);
		Assert.assertEquals(tmp[1], v.y);
		Assert.assertEquals(tmp[2], v.z);

		v = new Float3(tmp[0], tmp[1], tmp[2]);
		Assert.assertEquals(tmp[0], v.x);
		Assert.assertEquals(tmp[1], v.y);
		Assert.assertEquals(tmp[2], v.z);
	}

	@Test
	public void testCopy()
	{
		Float3 v1 = new Float3(mRandom.nextFloat(), mRandom.nextFloat(), mRandom.nextFloat());
		Float3 v2 = new Float3(v1);

		Assert.assertEquals(v2.x, v1.x);
		Assert.assertEquals(v2.y, v1.y);
		Assert.assertEquals(v2.z, v1.z);
	}

	@Test
	public void testMisc()
	{
		float[] tmp = new float[] { mRandom.nextFloat(), mRandom.nextFloat(), mRandom.nextFloat() };
		Float3 v = new Float3(tmp);

		float sqrLength = v.getSquaredLength();
		float expSqrLength = tmp[0] * tmp[0] + tmp[1] * tmp[1] + tmp[2] * tmp[2];
		Assert.assertEquals(expSqrLength, sqrLength);

		float length = v.getLength();
		float expLength = (float) Math.sqrt(expSqrLength);
		Assert.assertEquals(expLength, length);
	}

	@Test
	public void testMult()
	{
		float[] tmp = new float[] { mRandom.nextFloat(), mRandom.nextFloat(), mRandom.nextFloat() };
		float f = mRandom.nextFloat();

		Float3 v = new Float3(tmp);
		v.mult(f);
		Assert.assertEquals(tmp[0] * f, v.x);
		Assert.assertEquals(tmp[1] * f, v.y);
		Assert.assertEquals(tmp[2] * f, v.z);

		Float3 v1 = new Float3(tmp);
		Float3 v2 = Float3.mult(v1, f);

		Assert.assertEquals(tmp[0], v1.x);
		Assert.assertEquals(tmp[1], v1.y);
		Assert.assertEquals(tmp[2], v1.z);

		Assert.assertEquals(tmp[0] * f, v2.x);
		Assert.assertEquals(tmp[1] * f, v2.y);
		Assert.assertEquals(tmp[2] * f, v2.z);
	}

	@Test
	public void testSub()
	{
		float[] tmp1 = new float[] { mRandom.nextFloat(), mRandom.nextFloat(), mRandom.nextFloat() };
		float[] tmp2 = new float[] { mRandom.nextFloat(), mRandom.nextFloat(), mRandom.nextFloat() };

		Float3 v1 = new Float3(tmp1);
		Float3 v2 = new Float3(tmp2);

		Float3 v3 = Float3.sub(v1, v2);
		Assert.assertEquals(tmp1[0] - tmp2[0], v3.x);
		Assert.assertEquals(tmp1[1] - tmp2[1], v3.y);
		Assert.assertEquals(tmp1[2] - tmp2[2], v3.z);

		v1.sub(v2);
		Assert.assertEquals(tmp1[0] - tmp2[0], v1.x);
		Assert.assertEquals(tmp1[1] - tmp2[1], v1.y);
		Assert.assertEquals(tmp1[2] - tmp2[2], v1.z);
	}
}
